from pathlib import Path
import json
import warnings
from ._model_extension import ModelExtension

try:
    from fastai.vision import flatten_model, ImageList
    from fastai.vision import imagenet_stats
    import torch
    from fastai.torch_core import split_model_idx
    from .._utils.pascal_voc_rectangles import ObjectDetectionCategoryList
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from ._arcgis_model import _resnet_family, _get_device
    from ._timm_utils import filter_timm_models
    from .._utils.common import (
        get_multispectral_data_params_from_emd,
        _get_emd_path,
    )
    from ._arcgis_model import _resnet_family, _get_device
    from ._ssd_utils import AveragePrecision
    import types
    from torch.jit.annotations import List, Dict
    from torchvision.models.detection.roi_heads import fastrcnn_loss
    from torchvision.models.detection.transform import resize_boxes

    HAS_FASTAI = True

except Exception as e:
    HAS_FASTAI = False


class MyFasterRCNN:
    """
    Create class with following fixed function names and the number of arguents to train your model from external source
    """

    try:
        import torch
        import torchvision
        import fastai

        tvisver = [int(x) for x in torchvision.__version__.split(".")]
    except:
        pass

    def get_model(self, data, backbone=None, **kwargs):
        """
        In this fuction you have to define your model with following two arguments!

        data - Object returned from prepare_data method(Fastai databunch)

        These two arguments comes from dataset which you have prepared from prepare_data method above.

        """
        (
            self.fasterrcnn_kwargs,
            kwargs,
        ) = self.fastai.core.split_kwargs_by_func(
            kwargs, self.torchvision.models.detection.FasterRCNN.__init__
        )

        if "timm" in backbone:
            from arcgis.learn.models._timm_utils import timm_config, _get_feature_size

            backbone_cut = timm_config(backbone)["cut"]
        else:
            backbone_cut = None

        if backbone is None:
            backbone = self.torchvision.models.resnet50

        elif type(backbone) is str:
            if hasattr(self.torchvision.models, backbone):
                backbone = getattr(self.torchvision.models, backbone)
            elif hasattr(self.torchvision.models.detection, backbone):
                backbone = getattr(self.torchvision.models.detection, backbone)
            elif "timm:" in backbone:
                import timm

                bckbn = backbone.split(":")[1]
                if hasattr(timm.models, bckbn):
                    backbone = getattr(timm.models, bckbn)
        else:
            backbone = backbone
        pretrained_backbone = kwargs.get("pretrained_backbone", True)
        assert type(pretrained_backbone) == bool
        if backbone.__name__ == "resnet50" and "timm" not in backbone.__module__:
            model = self.torchvision.models.detection.fasterrcnn_resnet50_fpn(
                pretrained=pretrained_backbone,
                pretrained_backbone=False,
                min_size=1.5 * data.chip_size,
                max_size=2 * data.chip_size,
                **self.fasterrcnn_kwargs,
            )

        elif (
            backbone.__name__ in ["resnet101", "resnet152"]
            and "timm" not in backbone.__module__
        ):
            backbone_fpn = (
                self.torchvision.models.detection.backbone_utils.resnet_fpn_backbone(
                    backbone.__name__, pretrained=pretrained_backbone
                )
            )
            model = self.torchvision.models.detection.FasterRCNN(
                backbone_fpn,
                91,
                min_size=1.5 * data.chip_size,
                max_size=2 * data.chip_size,
                **self.fasterrcnn_kwargs,
            )

        else:
            backbone_small = self.fastai.vision.learner.create_body(
                backbone, pretrained_backbone, backbone_cut
            )
            if "timm" in backbone.__module__:
                from arcgis.learn.models._maskrcnn import TimmFPNBackbone

                try:
                    backbone_small = TimmFPNBackbone(backbone_small, data.chip_size)
                except:
                    pass

            if not hasattr(backbone_small, "out_channels"):
                if "tresnet" in backbone.__module__:
                    backbone_small.out_channels = _get_feature_size(
                        backbone, backbone_cut
                    )[-1][1]
                else:
                    backbone_small.out_channels = (
                        self.fastai.callbacks.hooks.num_features_model(
                            self.torch.nn.Sequential(*backbone_small.children())
                        )
                    )

            model = self.torchvision.models.detection.FasterRCNN(
                backbone_small,
                91,
                min_size=1.5 * data.chip_size,
                max_size=2 * data.chip_size,
                **self.fasterrcnn_kwargs,
            )

        in_features = model.roi_heads.box_predictor.cls_score.in_features
        model.roi_heads.box_predictor = (
            self.torchvision.models.detection.faster_rcnn.FastRCNNPredictor(
                in_features, len(data.classes)
            )
        )

        if data._is_multispectral:
            model.transform.image_mean = [0] * len(data._extract_bands)
            model.transform.image_std = [1] * len(data._extract_bands)

        model.roi_heads.nms_thresh = 0.1
        model.roi_heads.score_thresh = 0.2

        self.model = model

        return model

    def on_batch_begin(self, learn, model_input_batch, model_target_batch, **kwargs):
        """
        This fuction is dedicated to put the inputs and outputs of the model before training. This is equivalent to fastai
        on_batch_begin function. In this function you will get the inputs and targets with applied transormations. You should
        be very carefull to return the model input and target during traing, model will only accept model_input(in many cases it
        is possible to model accept input and target both to return the loss during traing and you don't require to compute loss
        from the model output and the target by yourself), if you want to compute the loss by yourself by taking the output of the
        model and targets then you have to return the model_target in desired format to calculate loss in the loss function.

        learn - Fastai learner object.
        model_input_batch - transformed input batch(images) with tensor shape [N,C,H,W].
        model_target_batch - transformed target batch. list with [bboxes, classes]. Where bboxes tensor shape will be
                            [N, maximum_num_of_boxes_pesent_in_one_image_of_the_batch, 4(y1,x1,y2,x2 fastai default bbox
                            formate)] and bboxes in the range from -1 to 1(default fastai formate), and classes is the tenosr
                            of shape [N, maximum_num_of_boxes_pesent_in_one_image_of_the_batch] which represents class of each
                            bboxes.
        if you are synthesizing new data from the model_target_batch and model_input_batch, in that case you need to put
        your data on correct device.

        return model_input and model_target from this function.

        """

        # during training after each epoch, validation loss is required on validation set of datset.
        # torchvision FasterRCNN model gives losses only on training mode that is why set your model in train mode
        # such that you can get losses for your validation datset as well after each epoch.
        train = kwargs.get("train")
        self.model.train()
        if train:
            self.model.roi_heads.train_val = False
            self.model.rpn.train_val = False
            self.model.train_val = False
            self.model.transform.train_val = False
        else:
            self.model.backbone.eval()  # to get feature in eval mode for evaluation
            self.model.roi_heads.train_val = True
            self.model.rpn.train_val = True
            self.model.train_val = True
            self.model.transform.train_val = True

        target_list = []

        # denormalize from imagenet_stats
        if not learn.data._is_multispectral:
            imagenet_stats = [[0.485, 0.456, 0.406], [0.229, 0.224, 0.225]]
            mean = self.torch.tensor(imagenet_stats[0], dtype=self.torch.float32).to(
                model_input_batch.device
            )
            std = self.torch.tensor(imagenet_stats[1], dtype=self.torch.float32).to(
                model_input_batch.device
            )
            model_input_batch = (
                model_input_batch.permute(0, 2, 3, 1) * std + mean
            ).permute(0, 3, 1, 2)

        for bbox, label in zip(*model_target_batch):
            bbox = (
                (bbox + 1) / 2
            ) * learn.data.chip_size  # FasterRCNN model require bboxes with values between 0 and H and 0 and W.
            mask = (bbox[:, 2:] >= (bbox[:, :2] + 1.0)).all(1)
            bbox = bbox[mask]
            label = label[mask]
            target = (
                {}
            )  # FasterRCNN require target of each image in the formate of dictionary.
            # If image comes without any bboxes.
            if (self.tvisver[0] == 0 and self.tvisver[1] < 6) and bbox.nelement() == 0:
                bbox = self.torch.tensor([[0.0, 0.0, 0.0, 0.0]]).to(learn.data.device)
                label = self.torch.tensor([0]).to(learn.data.device)
            # FasterRCNN require the formate of bboxes [x1,y1,x2,y2].
            bbox = self.torch.index_select(
                bbox,
                1,
                self.torch.tensor([1, 0, 3, 2]).to(learn.data.device),
            )
            target["boxes"] = bbox
            target["labels"] = label
            target_list.append(
                target
            )  # FasterRCNN require batches target in form of list of dictionary.

        # handle batch size one in training
        if model_input_batch.shape[0] < 2:
            model_input_batch = self.torch.cat((model_input_batch, model_input_batch))
            target_list.append(target_list[0])

        # FasterRCNN require model input with images and coresponding targets in training mode to return the losses so append
        # the targets in model input itself.
        model_input = [list(model_input_batch), target_list]
        # Model target is not required in traing mode so just return the same model_target to train the model.
        model_target = model_target_batch

        # return model_input and model_target
        return model_input, model_target

    def transform_input(self, xb, thresh=0.5, nms_overlap=0.1):  # transform_input
        """
        function for feding the input to the model in validation/infrencing mode.

        xb - tensor with shape [N, C, H, W]
        """
        self.model.roi_heads.train_val = False
        self.model.rpn.train_val = False
        self.model.train_val = False
        self.model.transform.train_val = False
        self.nms_thres = self.model.roi_heads.nms_thresh
        self.thresh = self.model.roi_heads.score_thresh
        self.model.roi_heads.nms_thresh = nms_overlap
        self.model.roi_heads.score_thresh = thresh

        # denormalize from imagenet_stats
        imagenet_stats = [[0.485, 0.456, 0.406], [0.229, 0.224, 0.225]]
        mean = self.torch.tensor(imagenet_stats[0], dtype=self.torch.float32).to(
            xb.device
        )
        std = self.torch.tensor(imagenet_stats[1], dtype=self.torch.float32).to(
            xb.device
        )

        xb = (xb.permute(0, 2, 3, 1) * std + mean).permute(0, 3, 1, 2)

        return list(xb)  # model input require in the formate of list

    def transform_input_multispectral(self, xb, thresh=0.5, nms_overlap=0.1):
        self.model.roi_heads.train_val = False
        self.model.rpn.train_val = False
        self.model.train_val = False
        self.model.transform.train_val = False
        self.nms_thres = self.model.roi_heads.nms_thresh
        self.thresh = self.model.roi_heads.score_thresh
        self.model.roi_heads.nms_thresh = nms_overlap
        self.model.roi_heads.score_thresh = thresh

        return list(xb)

    def loss(self, model_output, *model_target):
        """
        Define loss in this function.

        model_output - model output after feding input to the model in traing mode.
        *model_target - targets of the model which you have return in above on_batch_begin function.

        return loss for the model
        """
        if isinstance(model_output, tuple):
            model_output = model_output[1]
        # FasterRCNN model return loss in traing mode by feding input to the model it does not require target to compute the loss
        final_loss = 0.0
        for i in model_output.values():
            i[self.torch.isnan(i)] = 0.0
            i[self.torch.isinf(i)] = 0.0
            final_loss += i

        return final_loss

    def post_process(self, pred, nms_overlap, thres, chip_size, device):
        """
        Fuction dedicated for post processing your output of the model in validation/infrencing mode.

        pred - Predictions(output) of the model after feding the batch of input image.
        nms_overlap - If your model post processing require nms_overlap.
        thres - detction thresold if required in post processing.
        chip_size - If chip_size required in model post processing.
        device - device on which you should put you output after post processing.

        It should return the bboxes in range -1 to 1 and the formate of the post processed result is list of tuple for each
        image and tuple should contain (bboxes, label, score) for each image. bboxes should be the tensor of shape
        [Number_of_bboxes_in_image, 4], label should be the tensor of shape[Number_of_bboxes_in_image,] and score should be
        the tensor of shape[Number_of_bboxes_in_image,].
        """
        if not self.model.roi_heads.train_val:
            self.model.roi_heads.score_thresh = self.thresh
            self.model.roi_heads.nms_thresh = self.nms_thres

        post_processed_pred = []
        for p in pred:
            bbox, label, score = p["boxes"], p["labels"], p["scores"]
            # convert bboxes in range -1 to 1.
            bbox = bbox / (chip_size / 2) - 1
            # convert bboxes in format [y1,x1,y2,x2]
            bbox = self.torch.index_select(
                bbox, 1, self.torch.tensor([1, 0, 3, 2]).to(bbox.device)
            )
            # Append the tuple in list for each image
            post_processed_pred.append(
                (bbox.data.to(device), label.to(device), score.to(device))
            )

        return post_processed_pred


def forward_roi(self, features, proposals, image_shapes, targets=None):
    """
    Arguments:
        features (List[Tensor])
        proposals (List[Tensor[N, 4]])
        image_shapes (List[Tuple[H, W]])
        targets (List[Dict])
    """

    train_val = getattr(self, "train_val", False)

    if targets is not None:
        for t in targets:
            floating_point_types = (torch.float, torch.double, torch.half)
            assert (
                t["boxes"].dtype in floating_point_types
            ), "target boxes must of float type"
            assert t["labels"].dtype == torch.int64, "target labels must of int64 type"

    if self.training:
        if train_val:
            original_prpsl = [p.clone() for p in proposals]
        (
            proposals,
            matched_idxs,
            labels,
            regression_targets,
        ) = self.select_training_samples(proposals, targets)
    else:
        labels = None
        regression_targets = None
        matched_idxs = None

    box_features = self.box_roi_pool(features, proposals, image_shapes)
    box_features = self.box_head(box_features)
    class_logits, box_regression = self.box_predictor(box_features)

    result = torch.jit.annotate(List[Dict[str, torch.Tensor]], [])
    losses = {}
    if self.training:
        assert labels is not None and regression_targets is not None
        loss_classifier, loss_box_reg = fastrcnn_loss(
            class_logits, box_regression, labels, regression_targets
        )
        losses = {
            "loss_classifier": loss_classifier,
            "loss_box_reg": loss_box_reg,
        }
    if not self.training or train_val:
        if train_val:
            box_features = self.box_roi_pool(features, original_prpsl, image_shapes)
            box_features = self.box_head(box_features)
            class_logits, box_regression = self.box_predictor(box_features)
            boxes, scores, labels = self.postprocess_detections(
                class_logits, box_regression, original_prpsl, image_shapes
            )
        else:
            boxes, scores, labels = self.postprocess_detections(
                class_logits, box_regression, proposals, image_shapes
            )
        num_images = len(boxes)
        for i in range(num_images):
            result.append(
                {
                    "boxes": boxes[i],
                    "labels": labels[i],
                    "scores": scores[i],
                }
            )

    return result, losses


def postprocess_transform(self, result, image_shapes, original_image_sizes):
    train_val = getattr(self, "train_val", False)

    if not self.training or train_val:
        for i, (pred, im_s, o_im_s) in enumerate(
            zip(result, image_shapes, original_image_sizes)
        ):
            boxes = pred["boxes"]
            boxes = resize_boxes(boxes, im_s, o_im_s)
            result[i]["boxes"] = boxes

    elif self.training:
        return result

    return result


def post_nms_top_n(self):
    train_val = getattr(self, "train_val", False)

    if train_val:
        return self._post_nms_top_n["testing"]
    elif self.training:
        return self._post_nms_top_n["training"]
    return self._post_nms_top_n["testing"]


def pre_nms_top_n(self):
    train_val = getattr(self, "train_val", False)

    if train_val:
        self._pre_nms_top_n["testing"]
    elif self.training:
        return self._pre_nms_top_n["training"]
    return self._pre_nms_top_n["testing"]


if HAS_FASTAI:

    def eager_outputs_modified(self, losses, detections):
        train_val = getattr(self, "train_val", False)

        if train_val:
            return detections, losses
        elif self.training:
            return losses
        return detections


class FasterRCNN(ModelExtension):
    """
    Model architecture from https://arxiv.org/abs/1506.01497.
    Creates a ``FasterRCNN`` object detection model,
    based on https://github.com/pytorch/vision/blob/master/torchvision/models/detection/faster_rcnn.py.

    =============================   =============================================
    **Parameter**                    **Description**
    -----------------------------   ---------------------------------------------
    data                            Required fastai Databunch. Returned data object from
                                    :meth:`~arcgis.learn.prepare_data`  function.
    -----------------------------   ---------------------------------------------
    backbone                        Optional string. Backbone convolutional neural network
                                    model used for feature extraction, which
                                    is `resnet50` by default.
                                    Supported backbones: ResNet family and specified Timm
                                    models(experimental support) from :func:`~arcgis.learn.FasterRCNN.backbones`.
    -----------------------------   ---------------------------------------------
    pretrained_path                 Optional string. Path where pre-trained model is
                                    saved.
    =============================   =============================================

    **kwargs**

    =============================   =============================================
    **Parameter**                    **Description**
    -----------------------------   ---------------------------------------------
    rpn_pre_nms_top_n_train         Optional int. Number of proposals to keep before
                                    applying NMS during training.
                                    Default: 2000
    -----------------------------   ---------------------------------------------
    rpn_pre_nms_top_n_test          Optional int. Number of proposals to keep before
                                    applying NMS during testing.
                                    Default: 1000
    -----------------------------   ---------------------------------------------
    rpn_post_nms_top_n_train        Optional int. Number of proposals to keep after
                                    applying NMS during training.
                                    Default: 2000
    -----------------------------   ---------------------------------------------
    rpn_post_nms_top_n_test         Optional int. Number of proposals to keep after
                                    applying NMS during testing.
                                    Default: 1000
    -----------------------------   ---------------------------------------------
    rpn_nms_thresh                  Optional float. NMS threshold used for postprocessing
                                    the RPN proposals.
                                    Default: 0.7
    -----------------------------   ---------------------------------------------
    rpn_fg_iou_thresh               Optional float. Minimum IoU between the anchor
                                    and the GT box so that they can be considered
                                    as positive during training of the RPN.
                                    Default: 0.7
    -----------------------------   ---------------------------------------------
    rpn_bg_iou_thresh               Optional float. Maximum IoU between the anchor and
                                    the GT box so that they can be considered as negative
                                    during training of the RPN.
                                    Default: 0.3
    -----------------------------   ---------------------------------------------
    rpn_batch_size_per_image        Optional int. Number of anchors that are sampled
                                    during training of the RPN for computing the loss.
                                    Default: 256
    -----------------------------   ---------------------------------------------
    rpn_positive_fraction           Optional float. Proportion of positive anchors in a
                                    mini-batch during training of the RPN.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_score_thresh                Optional float. During inference, only return proposals
                                    with a classification score greater than box_score_thresh
                                    Default: 0.05
    -----------------------------   ---------------------------------------------
    box_nms_thresh                  Optional float. NMS threshold for the prediction head.
                                    Used during inference.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_detections_per_img          Optional int. Maximum number of detections per
                                    image, for all classes.
                                    Default: 100
    -----------------------------   ---------------------------------------------
    box_fg_iou_thresh               Optional float. Minimum IoU between the proposals and
                                    the GT box so that they can be considered as positive
                                    during training of the classification head.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_bg_iou_thresh               Optional float. Maximum IoU between the proposals and
                                    the GT box so that they can be considered as negative
                                    during training of the classification head.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_batch_size_per_image        Optional int. Number of proposals that are sampled during
                                    training of the classification head.
                                    Default: 512
    -----------------------------   ---------------------------------------------
    box_positive_fraction           Optional float. Proportion of positive proposals in a
                                    mini-batch during training of the classification head.
                                    Default: 0.25
    =============================   =============================================

    :return:
        :class:`~arcgis.learn.FasterRCNN` Object

    """

    def __init__(self, data, backbone="resnet50", pretrained_path=None, **kwargs):
        self._check_dataset_support(data)
        backbone_name = backbone if type(backbone) is str else backbone.__name__
        if backbone_name not in self.supported_backbones:
            raise Exception(
                f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"
            )

        super().__init__(data, MyFasterRCNN, backbone, pretrained_path, **kwargs)

        self.learn.model.roi_heads.forward = types.MethodType(
            forward_roi, self.learn.model.roi_heads
        )
        self.learn.model.eager_outputs = types.MethodType(
            eager_outputs_modified, self.learn.model
        )
        self.learn.model.transform.postprocess = types.MethodType(
            postprocess_transform, self.learn.model.transform
        )
        self.learn.model.rpn.post_nms_top_n = types.MethodType(
            post_nms_top_n, self.learn.model.rpn
        )
        self.learn.model.rpn.pre_nms_top_n = types.MethodType(
            pre_nms_top_n, self.learn.model.rpn
        )
        self.learn.metrics = [AveragePrecision(self, data.c - 1)]
        idx = self._freeze()
        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])
        self.learn.create_opt(lr=3e-3)

    def unfreeze(self):
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    def _freeze(self):
        "Freezes the pretrained backbone."
        for idx, i in enumerate(flatten_model(self.learn.model.backbone)):
            if isinstance(i, (torch.nn.BatchNorm2d)):
                continue
            for p in i.parameters():
                p.requires_grad = False
        return idx

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "average_precision"]

    @property
    def _is_fasterrcnn(self):
        return True

    @property
    def supported_backbones(self):
        """Supported list of backbones for this model."""
        return FasterRCNN._supported_backbones()

    @staticmethod
    def backbones():
        """Supported list of backbones for this model."""
        return FasterRCNN._supported_backbones()

    @staticmethod
    def _supported_backbones():
        timm_models = filter_timm_models(["*repvgg*", "*tresnet*"])
        timm_backbones = list(map(lambda m: "timm:" + m, timm_models))
        return [*_resnet_family] + timm_backbones

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return FasterRCNN._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["PASCAL_VOC_rectangles", "KITTI_rectangles"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a ``FasterRCNN`` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return:
            :class:`~arcgis.learn.FasterRCNN` Object
        """
        emd_path = _get_emd_path(emd_path)

        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        backbone = emd["ModelParameters"]["backbone"]
        dataset_type = emd.get("DatasetType", "PASCAL_VOC_rectangles")
        chip_size = emd["ImageWidth"]
        resize_to = emd.get("resize_to", None)
        kwargs = emd.get("Kwargs", {})
        if isinstance(resize_to, list):
            resize_to = (resize_to[0], resize_to[1])

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        data_passed = True
        if data is None:
            data_passed = False
            train_tfms = []
            val_tfms = []
            ds_tfms = (train_tfms, val_tfms)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                sd = ImageList([], path=emd_path.parent.parent.parent).split_by_idx([])
                data = (
                    sd.label_const(
                        0,
                        label_cls=ObjectDetectionCategoryList,
                        classes=list(class_mapping.values()),
                    )
                    .transform(ds_tfms)
                    .databunch(device=_get_device())
                    .normalize(imagenet_stats)
                )
            # Add 1 for background class
            data.c += 1
            data.chip_size = chip_size
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data.classes = ["background"] + list(class_mapping.values())
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)
            data.dataset_type = dataset_type

        data.resize_to = resize_to
        frcnn = cls(data, backbone, pretrained_path=str(model_file), **kwargs)

        if not data_passed:
            frcnn.learn.data.single_ds.classes = frcnn._data.classes
            frcnn.learn.data.single_ds.y.classes = frcnn._data.classes

        return frcnn

    def predict(
        self,
        image_path,
        threshold=0.5,
        nms_overlap=0.1,
        return_scores=False,
        visualize=False,
        resize=False,
    ):
        """
        Runs prediction on an Image. This method is only supported for RGB images.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean. Will return the probability
                                scores of the bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with
                                predicted bounding boxes if True.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the
                                same size (chip_size parameter in prepare_data)
                                that the model was trained on, before detecting
                                objects. Note that if resize_to parameter was
                                used in prepare_data, the image is resized to
                                that size instead.

                                By default, this parameter is false and the
                                detections are run in a sliding window fashion
                                by applying the model on cropped sections of
                                the image (of the same size as the model was
                                trained on).
        =====================   ===========================================

        :return: Returns a tuple with predictions, labels and optionally confidence scores
                  if return_scores=True. The predicted bounding boxes are returned as a list
                  of lists containing the  xmin, ymin, width and height of each predicted object
                  in each image. The labels are returned as a list of class values and the
                  confidence scores are returned as a list of floats indicating the confidence
                  of each prediction.
        """

    def predict_video(
        self,
        input_video_path,
        metadata_file,
        threshold=0.5,
        nms_overlap=0.1,
        track=False,
        visualize=False,
        output_file_path=None,
        multiplex=False,
        multiplex_file_path=None,
        tracker_options={
            "assignment_iou_thrd": 0.3,
            "vanish_frames": 40,
            "detect_frames": 10,
        },
        visual_options={
            "show_scores": True,
            "show_labels": True,
            "thickness": 2,
            "fontface": 0,
            "color": (255, 255, 255),
        },
        resize=False,
    ):
        """
        Runs prediction on a video and appends the output VMTI predictions in the metadata file.
        This method is only supported for RGB images.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        input_video_path        Required. Path to the video file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        metadata_file           Required. Path to the metadata csv file where
                                the predictions will be saved in VMTI format.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        track                   Optional bool. Set this parameter as True to
                                enable object tracking.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. If True a video is saved
                                with prediction results.
        ---------------------   -------------------------------------------
        output_file_path        Optional path. Path of the final video to be saved.
                                If not supplied, video will be saved at path input_video_path
                                appended with _prediction.
        ---------------------   -------------------------------------------
        multiplex               Optional boolean. Runs Multiplex using the VMTI detections.
        ---------------------   -------------------------------------------
        multiplex_file_path     Optional path. Path of the multiplexed video to be saved.
                                By default a new file with _multiplex.MOV extension is saved
                                in the same folder.
        ---------------------   -------------------------------------------
        tracking_options        Optional dictionary. Set different parameters for
                                object tracking. assignment_iou_thrd parameter is used
                                to assign threshold for assignment of trackers,
                                vanish_frames is the number of frames the object should
                                be absent to consider it as vanished, detect_frames
                                is the number of frames an object should be detected
                                to track it.
        ---------------------   -------------------------------------------
        visual_options          Optional dictionary. Set different parameters for
                                visualization.
                                show_scores boolean, to view scores on predictions,
                                show_labels boolean, to view labels on predictions,
                                thickness integer, to set the thickness level of box,
                                fontface integer, fontface value from opencv values,
                                color tuple (B, G, R), tuple containing values between
                                0-255.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the video frames to the same size
                                (chip_size parameter in prepare_data) that the model was
                                trained on, before detecting objects. Note that if
                                resize_to parameter was used in prepare_data,
                                the video frames are resized to that size instead.

                                By default, this parameter is false and the detections
                                are run in a sliding window fashion by applying the
                                model on cropped sections of the frame (of the same
                                size as the model was trained on).
        =====================   ===========================================

        """

    def average_precision_score(
        self,
        detect_thresh=0.2,
        iou_thresh=0.1,
        mean=False,
        show_progress=True,
    ):
        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """

    def show_results(self, rows=5, thresh=0.5, nms_overlap=0.1):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        =====================   ===========================================

        """



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model settings
model = dict(
    type="GridRCNN",
    pretrained="torchvision://resnet50",
    backbone=dict(
        type="ResNet",
        depth=50,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        norm_eval=True,
        style="pytorch",
    ),
    neck=dict(
        type="FPN", in_channels=[256, 512, 1024, 2048], out_channels=256, num_outs=5
    ),
    rpn_head=dict(
        type="RPNHead",
        in_channels=256,
        feat_channels=256,
        anchor_generator=dict(
            type="AnchorGenerator",
            scales=[8],
            ratios=[0.5, 1.0, 2.0],
            strides=[4, 8, 16, 32, 64],
        ),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=[0.0, 0.0, 0.0, 0.0],
            target_stds=[1.0, 1.0, 1.0, 1.0],
        ),
        loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0),
        loss_bbox=dict(type="SmoothL1Loss", beta=1.0 / 9.0, loss_weight=1.0),
    ),
    roi_head=dict(
        type="GridRoIHead",
        bbox_roi_extractor=dict(
            type="SingleRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=7, sampling_ratio=0),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32],
        ),
        bbox_head=dict(
            type="Shared2FCBBoxHead",
            with_reg=False,
            in_channels=256,
            fc_out_channels=1024,
            roi_feat_size=7,
            num_classes=80,
            bbox_coder=dict(
                type="DeltaXYWHBBoxCoder",
                target_means=[0.0, 0.0, 0.0, 0.0],
                target_stds=[0.1, 0.1, 0.2, 0.2],
            ),
            reg_class_agnostic=False,
        ),
        grid_roi_extractor=dict(
            type="SingleRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=14, sampling_ratio=0),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32],
        ),
        grid_head=dict(
            type="GridHead",
            grid_points=9,
            num_convs=8,
            in_channels=256,
            point_feat_channels=64,
            norm_cfg=dict(type="GN", num_groups=36),
            loss_grid=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=15),
        ),
    ),
    # model training and testing settings
    train_cfg=dict(
        rpn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.7,
                neg_iou_thr=0.3,
                min_pos_iou=0.3,
                ignore_iof_thr=-1,
            ),
            sampler=dict(
                type="RandomSampler",
                num=256,
                pos_fraction=0.5,
                neg_pos_ub=-1,
                add_gt_as_proposals=False,
            ),
            allowed_border=0,
            pos_weight=-1,
            debug=False,
        ),
        rpn_proposal=dict(
            nms_across_levels=False,
            nms_pre=2000,
            nms_post=2000,
            max_num=2000,
            nms_thr=0.7,
            min_bbox_size=0,
        ),
        rcnn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.5,
                neg_iou_thr=0.5,
                min_pos_iou=0.5,
                ignore_iof_thr=-1,
            ),
            sampler=dict(
                type="RandomSampler",
                num=512,
                pos_fraction=0.25,
                neg_pos_ub=-1,
                add_gt_as_proposals=True,
            ),
            pos_radius=1,
            pos_weight=-1,
            max_num_grid=192,
            debug=False,
        ),
    ),
    test_cfg=dict(
        rpn=dict(
            nms_across_levels=False,
            nms_pre=1000,
            nms_post=1000,
            max_num=1000,
            nms_thr=0.7,
            min_bbox_size=0,
        ),
        rcnn=dict(
            score_thr=0.03, nms=dict(type="nms", iou_threshold=0.3), max_per_img=100
        ),
    ),
)



from ._arcgis_model import ArcGISModel
from pathlib import Path
import json
import random
import traceback
from ._arcgis_model import _EmptyData
from .._utils.env import raise_fastai_import_error
import copy

try:
    import torch
    import glob
    from torch import nn
    from .._utils.common import _get_emd_path
    from ._siammask_utils import (
        train_callback,
        get_learner,
        Custom,
        load_pretrain,
        siamese_init,
        siamese_track,
        calculate_iou,
        download_backbone,
    )
    import matplotlib.pyplot as plt
    from fastai.torch_core import split_model_idx
    from fastai.vision import flatten_model

    from fastai.basic_train import Learner
    from fastai.vision import ImageList
    from .._utils.pascal_voc_rectangles import ObjectDetectionCategoryList

    import numpy as np
    import cv2, os
    from types import SimpleNamespace
    from os import makedirs
    from os.path import join, isdir, isfile

    from torch.autograd import Variable
    import torch.nn.functional as F
    from .._utils.object_tracking_data import get_image_for_tracking

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class Track:
    """
    Creates a Track object, used to maintain the state of a track

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    id                      Required int. ID for each track initialized
    ---------------------   -------------------------------------------
    label                   Required String. label/class name of the track
    ---------------------   -------------------------------------------
    bbox                    Required list. Bounding box of the track
    ---------------------   -------------------------------------------
    mask                    Required numpy array. Mask for the tack
    =====================   ===========================================

    :return: :class:`~arcgis.learn.Track` Object
    """

    def __init__(self, id, label, bbox, mask):
        self.id = id
        self.label = label
        self.bbox = bbox
        self.score = 1
        self.status = 16
        self.mask = mask
        self.location = None
        self.age = 0
        # self.color = color


class SiamMask(ArcGISModel):
    """
    Creates a :class:`~arcgis.learn.SiamMask` object.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Optional fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data` function with dataset_type as
                            'ObjectTracking' and data format as 'YouTube-VOS'.
                            Default value is None.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.SiamMask` Object
    """

    def __init__(self, data=None, **kwargs):
        if not HAS_FASTAI:
            raise_fastai_import_error(
                import_exception=import_exception, message="", installation_steps=" "
            )

        self._is_multispectral = False

        pretrained_path = kwargs.get("pretrained_path", None)
        self.cfg = {}
        self.anchors = {
            "stride": 8,
            "ratios": [0.33, 0.5, 1, 2, 3],
            "scales": [8],
            "round_dight": 0,
        }
        self.cfg["network"] = {"arch": "Custom"}
        self.cfg["hp"] = {
            "instance_size": 255,
            "base_size": 8,
            "out_size": 127,
            "seg_thr": 0.35,
            "penalty_k": 0.04,
            "window_influence": 0.4,
            "lr": 1.0,
        }

        self.track_list = []
        self.state_list = {}
        self.num_tracks = 0

        if data is not None:
            super().__init__(data)
            self.learn = get_learner(data=data, anchors=self.anchors)
            self.learn.callbacks.append(train_callback(self.learn))
        else:
            data = create_siammask_data()

            file_path = download_backbone(
                url="http://www.robots.ox.ac.uk/~qwang/SiamMask_DAVIS.pth",
                file_name="SiamMask_DAVIS.pth",
            )
            model = Custom(anchors=self.anchors, pretrain=False)

            model = load_pretrain(model, file_path)

            if pretrained_path is not None:
                self.load(pretrained_path)

            self.learn = Learner(data=data, model=model)
            self._backend = "pytorch"
            self._data = data
            self._learning_rate = None
            self._model_metrics_cache = None

            class Resnet50:
                def __init__(self):
                    self.name = "Resnet50"

            self._backbone = Resnet50

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        # self.learn.model = self.learn.model.to(self._device)
        # self.learn.c_device = self._device

        self.freeze()
        self._arcgis_init_callback()

        self.load_model = pretrained_path
        self._model = None
        if self.load_model is not None:
            if not os.path.isfile(self.load_model):
                raise Exception("Please check model file path")
            try:
                pretrained_path = self.load_model
                siammask = Custom(anchors=self.anchors)
                self._model = load_pretrain(siammask, pretrained_path)
                self.load(pretrained_path)
            except Exception:
                raise Exception("Error loading model")
        else:
            self._model = self.learn.model

        self.iou_threshold = 0

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "mIOU"]

    @staticmethod
    def _supported_backbones():
        """Supported torchvision backbones for this model."""
        return ["resnet50"]

    @property
    def supported_backbones(self):
        """Supported torchvision backbones for this model."""
        return SiamMask._supported_backbones()

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.SiamMask` Object tracker from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.SiamMask` Object
        """
        emd_path = _get_emd_path(emd_path)
        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        if data is None:
            data = _EmptyData(path=emd_path.parent, loss_func=None, c=2, chip_size=224)

            data.emd_path = emd_path
            data.emd = emd
            data._dataset_type = "_emptydata"
            for key, value in emd["DataAttributes"].items():
                setattr(data, key, value)

        return cls(
            data, **model_params, pretrained_path=str(model_file), load_from_model=True
        )

    def _get_emd_params(self, save_inference_file):
        _emd_template = {"DataAttributes": {}, "ModelParameters": {}}
        # chip size
        _emd_template["DataAttributes"]["chip_size"] = self._data.chip_size
        _emd_template["DataAttributes"][
            "_is_multispectral"
        ] = self._data._is_multispectral

        norm_stats = []
        for k in self._data.infos:
            norm_stats.append(k)
        _emd_template["DataAttributes"]["norm_stats"] = list(norm_stats)
        _emd_template["DataAttributes"]["batch_size"] = self._data.batch_size
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        # object classifier config can be used.
        _emd_template["ModelConfiguration"] = "_siammask_inferencing"
        _emd_template["ModelParameters"]["hp"] = self.cfg["hp"]
        _emd_template["ModelParameters"]["anchors"] = self.anchors
        # Model Type
        _emd_template["ModelType"] = "ObjectTracking"

        return _emd_template

    def freeze(self):
        "Freezes the pretrained backbone."

        self.learn.layer_groups = split_model_idx(self.learn.model, [158])
        self.learn.create_opt(lr=0.000478630092322)

    def init(self, frame, detections, labels=None, reset=True, **kwargs):
        """
        Initializes the position of the object in the frame/Image using detections.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        frame                   Required numpy array. frame is used to
                                initialize the objects to track.
        ---------------------   -------------------------------------------
        detections              Required list. A list of bounding boxes.
        ---------------------   -------------------------------------------
        labels                  Optional list. A list of labels corresponding
                                to the bounding boxes.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.Track` list
        """
        if self._model is None:
            raise Exception("SiamMask model not loaded properly.")

        if reset:
            self.track_list = []
            self.state_list = {}
            self.num_tracks = 0

        siammask = copy.deepcopy(self._model)
        siammask.eval().to(self.device)

        if detections is None:
            detections = []

        filtered_detections = []
        for i, detection in enumerate(detections):
            x, y = detection[0], detection[1]
            w, h = detection[2], detection[3]
            add = True
            for key, val in self.state_list.items():
                v = val[1]

                box1 = v[0], v[1], v[2], v[3]
                box2 = x, y, w, h

                iou = calculate_iou(box1, box2)
                if iou > self.iou_threshold:
                    add = False
                    break
            if add:
                filtered_detections.append(detection)
                # x, y = detection[0], detection[1]
                # w, h = detection[2], detection[3]
                target_pos = np.array([x + w / 2, y + h / 2])
                target_sz = np.array([w, h])
                state = siamese_init(
                    frame,
                    target_pos,
                    target_sz,
                    siammask,
                    self.cfg["hp"],
                    device=self.device,
                )
                # mask = state['mask'] > state['p'].seg_thr

                if labels is not None and len(labels) == len(detections):
                    track = Track(self.num_tracks, labels[i], [x, y, w, h], None)
                else:
                    track = Track(self.num_tracks, "Object", [x, y, w, h], None)
                self.track_list.append(track)
                self.state_list[track.id] = [state, [x, y, w, h]]
                self.num_tracks += 1

        return self.track_list

    def update(self, frame, **kwargs):
        """
        Tracks the position of the object in the frame/Image

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        frame                   Required numpy array. frame is used to update
                                the object track.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        detections              Optional list. A list of bounding boxes.
        ---------------------   -------------------------------------------
        labels                  Optional list. A list of labels.
        =====================   ===========================================

        :return: Updated track list
        """
        for i, track in enumerate(self.track_list):
            state = siamese_track(
                self.state_list[track.id][0],
                frame,
                mask_enable=True,
                refine_enable=True,
                device=self.device,
            )
            location = state["ploygon"].flatten()
            mask = state["mask"] > state["p"].seg_thr
            # frame[:, :, 2] = (mask > 0) * 255 + (mask == 0) * frame[:, :, 2]
            target_pos = state["target_pos"]
            w, h = state["target_sz"]
            x, y = target_pos[0] - (w / 2), target_pos[1] - (h / 2)
            self.track_list[i].bbox = [x, y, w, h]
            self.track_list[i].mask = mask
            self.track_list[i].location = location
            self.track_list[i].age += 1
            self.track_list[i].label = track.label
            self.track_list[i].score = state["score"]
            self.state_list[track.id][1] = [x, y, w, h]

        detections = kwargs.get("detections", None)
        if detections:
            labels = kwargs.get("labels", None)
            self.track_list = self.init(
                frame, detections=detections, labels=labels, reset=False
            )
        return self.track_list

    def remove(self, track_ids):
        """
        Removes the tracks from the track list using track_ids

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        track_ids               Required List. List of track ids to be removed
                                from the track list.
        =====================   ===========================================

        :return: Updated track list
        """
        try:
            for id in track_ids:
                for i in self.track_list:
                    if i.id == id:
                        self.track_list.remove(i)
                try:
                    del self.state_list[id]
                except Exception as e:
                    print(e)

                # TODO: usage of num_tracks needs to be revisited
                # if self.num_tracks > 0:
                #    self.num_tracks -= 1
            # print("Tracks has been removed succesfully!")
        except Exception as e:
            print(e)

        return self.track_list

    def show_results(self, rows=5):
        """
        Displays the results of a trained model on a part of the validation set

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows to display.
        =====================   ===========================================
        """
        if self._data._dataset_type == "_ObjectTracking":
            self._check_requisites()
            return None
        self._check_requisites()
        track_list_copy = copy.deepcopy(self.track_list)
        state_list_copy = self.state_list
        num_tracks_copy = self.num_tracks

        self.track_list = []
        self.state_list = {}
        self.num_tracks = 0

        idx = random.randint(0, len(self.learn.data.valid_ds) - 1)
        data = self.learn.data.valid_ds.__getitem__(
            idx, debug=False, show_batch=False, show_results=True
        )
        base_folder = data[0].split("crop")[0]
        folder_name = data[0].split("\\")[-2]
        frame_id = data[0].split("\\")[-1].split(".")[1]
        json_data = self.learn.data.valid_ds.show_batch_data[folder_name][frame_id]
        bbox = json_data[list(json_data.keys())[0]]

        all_images = []
        folder_path = os.path.join(base_folder, "JPEGImages", folder_name)
        all_images = glob.glob(
            os.path.join(base_folder, "JPEGImages", folder_name, "*.jpg")
        )
        if not os.path.isdir(folder_path):
            folder_path = os.path.join(base_folder, "images")
            combined_images = [
                file
                for file in os.listdir(folder_path)
                if file.endswith(".png")
                or file.endswith(".jpg")
                or file.endswith(".tif")
            ]
            if len(combined_images) > 0:
                ext = os.path.splitext(combined_images[0])[-1]
                if ext != ".tif":
                    ext = ".png"

                all_images = [
                    os.path.join(folder_path, file)
                    for file in combined_images
                    if os.path.isfile(
                        os.path.join(
                            base_folder,
                            "labels",
                            folder_name,
                            os.path.splitext(file)[0] + ext,
                        )
                    )
                ]

        image_counter = 0
        track_id_set = set()

        if (len(all_images) // 3) < rows:
            rows = len(all_images) // 3

        if rows <= 0:
            rows = 1

        fig, axes = plt.subplots(
            nrows=rows, ncols=3, squeeze=False, figsize=(20, rows * 3)
        )

        for idx in range(0, rows):
            if len(all_images) == image_counter:
                break

            for j in range(0, 3):
                if len(all_images) == image_counter:
                    break

                if image_counter == 0:
                    img_path = all_images[image_counter]
                    image = get_image_for_tracking(img_path)
                    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                    x, y = int(bbox[0]), int(bbox[1])
                    x1, y1 = int(bbox[2]) - int(bbox[0]), int(bbox[3]) - int(bbox[1])
                    w, h = x1 - x, y1 - y
                    detections = [[x, y, w, h]]
                    cv2.rectangle(
                        image,
                        (int(bbox[0]), int(bbox[1])),
                        (int(bbox[2]) - int(bbox[0]), int(bbox[3] - int(bbox[1]))),
                        (255, 0, 0),
                        2,
                    )

                    self.init(image, detections)
                else:
                    img_path = all_images[image_counter]
                    image = get_image_for_tracking(img_path)
                    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                    tracks = self.update(image)
                    for track in tracks:
                        track_id_set.add(track.id)
                        mask = track.mask
                        image[:, :, 2] = (mask > 0) * 255 + (mask == 0) * image[:, :, 2]
                        cv2.polylines(
                            image,
                            [np.int0(track.location).reshape((-1, 1, 2))],
                            True,
                            (w, 255, h),
                            3,
                        )

                axes[idx][j].set_xticks([])
                axes[idx][j].set_yticks([])
                axes[idx][j].imshow(image)
                image_counter += 1

        self.track_list = track_list_copy
        self.state_list = state_list_copy
        self.num_tracks = num_tracks_copy

    @property
    def _model_metrics(self):
        if self._data._dataset_type == "_ObjectTracking":
            return {}
        self._check_requisites()
        return self.compute_metrics()

    def compute_metrics(self, iou_thres=0.2):
        """
        Computes mean IOU and f-measure on validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth mask, above
                                which a predicted mask will be
                                considered a true positive.
        =====================   ===========================================

        :return: `dict` with mean IOU and F-Measure
        """
        self._check_requisites()
        if self._data._dataset_type == "_ObjectTracking":
            return {}
        tp = 0
        fp = 0
        fn = 0
        all_ious = []
        data = self.learn.data
        sep = os.sep
        for folder in data.val_folders.keys():
            is_pro_data = False
            all_images = glob.glob(
                os.path.join(data.path, "JPEGImages", folder, "*.jpg")
            )

            if len(all_images) == 0:
                for ext in [".jpg", ".png", ".tif"]:
                    all_images = glob.glob(os.path.join(data.path, "images", "*" + ext))
                    if len(all_images) != 0:
                        is_pro_data = True
                        break

            init = False
            for img in all_images:
                anno, ext = os.path.splitext((os.path.basename(img)))
                if ext != ".tif":
                    ext = ".png"

                if not is_pro_data:
                    anno = img.replace("JPEGImages", "Annotations").replace(
                        "jpg", "png"
                    )
                else:
                    anno = os.path.join(data.path, "labels", folder, anno + ext)

                if not os.path.isfile(anno):
                    continue

                anno_img = get_image_for_tracking(anno, grayscale=True)
                if len(np.unique(anno_img)) < 2:
                    continue

                objects = data.val_folders[folder].keys()
                img_name = img.split(sep)[-1].split(".")[0]
                gt_bboxes = []
                for obj in objects:
                    img_list = [x for x in list(data.val_folders[folder][obj].keys())]
                    for i in img_list:
                        if int(img_name) == int(i):
                            gt_bbox = data.val_folders[folder][obj][i]
                            gt_bboxes.append(gt_bbox)
                            break

                image = get_image_for_tracking(img)
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

                detections = []
                for bbox in gt_bboxes:
                    x, y = int(bbox[0]), int(bbox[1])
                    x1, y1 = int(bbox[2]) - int(bbox[0]), int(bbox[3]) - int(bbox[1])
                    w, h = x1 - x, y1 - y
                    detections.append([x, y, w, h])

                if not init:
                    self.init(image, detections)
                    init = True
                    clr = np.unique(anno_img)
                else:
                    tracks = self.update(image)
                    fn_temp = 0
                    tp_temp = 0
                    fp_temp = 0
                    for ind, track in enumerate(tracks):
                        mask = track.mask
                        if mask.shape[0] == 0:
                            fn_temp += 1
                            continue

                        pred_mask = np.zeros(
                            (image.shape[0], image.shape[1], 3), np.uint8
                        )
                        pred_mask[:, :, 2] = (mask > 0) * 255 + (mask == 0) * pred_mask[
                            :, :, 2
                        ]
                        pred_mask = cv2.cvtColor(pred_mask, cv2.COLOR_BGR2GRAY)
                        pred_mask[np.where(pred_mask > 0)] = 255

                        anno_mask = (anno_img == clr[ind + 1]).astype(np.uint8)
                        contour, _ = cv2.findContours(
                            anno_mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE
                        )
                        gt_mask = np.zeros(
                            (image.shape[0], image.shape[1], 3), np.uint8
                        )
                        gt_mask = cv2.drawContours(
                            gt_mask, contour, -1, (255, 255, 255), -1
                        )

                        gt_mask = cv2.cvtColor(gt_mask, cv2.COLOR_BGR2GRAY)

                        iou = get_ious(pred_mask, gt_mask)

                        if iou >= iou_thres:
                            tp_temp += 1
                        else:
                            fp_temp += 1

                        all_ious.append(iou)
                    fn += fn_temp
                    tp += tp_temp
                    fp += fp_temp

        f_measure = get_f_measure(tp, fp, fn)
        mean_iou = np.mean(all_ious, axis=0)

        self.track_list = []
        self.state_list = {}
        self.num_tracks = 0

        return {"mean_IOU": mean_iou, "f_measure": f_measure}


def create_siammask_data():
    """Create an empty databunch for siammask dataset."""

    train_tfms = []
    val_tfms = []
    ds_tfms = (train_tfms, val_tfms)

    siammask_class_ids = [0]

    siammask_label_names = "background"

    siammask_class_mapping = {k: v for k, v in enumerate(siammask_label_names)}
    class_mapping = {
        k: v for k, v in siammask_class_mapping.items() if k in siammask_class_ids
    }

    import tempfile

    sd = ImageList(
        [], path=tempfile.NamedTemporaryFile().name, ignore_empty=True
    ).split_none()
    data = (
        sd.label_const(
            0,
            label_cls=ObjectDetectionCategoryList,
            classes=list(class_mapping.values()),
        )
        .transform(ds_tfms)
        .databunch()
    )

    data.class_mapping = class_mapping
    data.classes = list(class_mapping.values())
    data._is_empty = False
    data._is_siammask = True
    data.resize_to = 127
    data.chip_size = 127
    data._is_multispectral = False

    data.infos = {"template": 127, "search": 143}
    data._dataset_type = "_ObjectTracking"
    # data._is_empty = True

    return data


def get_ious(pred_mask, gt_mask):
    mask_sum = (pred_mask > 0).astype(np.uint8) + (gt_mask > 0).astype(np.uint8)
    intxn = np.sum(mask_sum == 2)
    union = np.sum(mask_sum > 0)
    iou = intxn / (union + 1e-6)
    return iou


def get_f_measure(tp, fp, fn):
    precision = tp / (tp + fp + 1e-6)
    recall = tp / (tp + fn + 1e-6)
    if precision > 0:
        f_measure = (2 * precision * recall) / (precision + recall)
    else:
        f_measure = 0.0
    return f_measure



import os
import copy
import torch
import numpy as np
import pandas as pd
import torch.nn as nn
from fastai.callback import Callback
from fastai.torch_core import add_metrics
import torch.nn.init as init
import torch.nn.functional as F
from torch.autograd import Variable
from torch.utils.data import DataLoader
from sklearn.metrics import confusion_matrix
from fastprogress.fastprogress import progress_bar


def get_decoder(n_neurons):
    layers = []
    for i in range(len(n_neurons) - 1):
        layers.append(nn.Linear(n_neurons[i], n_neurons[i + 1]))
        if i < (len(n_neurons) - 2):
            layers.extend([nn.BatchNorm1d(n_neurons[i + 1]), nn.ReLU()])
    m = nn.Sequential(*layers)
    return m


"""
Pixel-Set encoder module
author: Vivien Sainte Fare Garnot
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import copy


class PixelSetEncoder(nn.Module):
    def __init__(self, input_dim, mlp1, pooling, mlp2, with_extra, extra_size):
        """
        Pixel-set encoder.
        Args:
            input_dim (int): Number of channels of the input tensors
            mlp1 (list):  Dimensions of the successive feature spaces of MLP1
            pooling (str): Pixel-embedding pooling strategy, can be chosen in ('mean','std','max,'min')
                or any underscore-separated combination thereof.
            mlp2 (list): Dimensions of the successive feature spaces of MLP2
            with_extra (bool): Whether additional pre-computed features are passed between the two MLPs
            extra_size (int, optional): Number of channels of the additional features, if any.
        """

        super(PixelSetEncoder, self).__init__()

        self.input_dim = input_dim
        self.mlp1_dim = copy.deepcopy(mlp1)
        self.mlp2_dim = copy.deepcopy(mlp2)
        self.pooling = pooling

        self.with_extra = with_extra
        self.extra_size = extra_size

        self.name = "PSE-{}-{}-{}".format(
            "|".join(list(map(str, self.mlp1_dim))),
            pooling,
            "|".join(list(map(str, self.mlp2_dim))),
        )

        self.output_dim = (
            input_dim * len(pooling.split("_"))
            if len(self.mlp2_dim) == 0
            else self.mlp2_dim[-1]
        )

        inter_dim = self.mlp1_dim[-1] * len(pooling.split("_"))

        if self.with_extra:
            self.name += "Extra"
            inter_dim += self.extra_size

        assert input_dim == mlp1[0]
        assert inter_dim == mlp2[0]
        # Feature extraction
        layers = []
        for i in range(len(self.mlp1_dim) - 1):
            layers.append(linlayer(self.mlp1_dim[i], self.mlp1_dim[i + 1]))
        self.mlp1 = nn.Sequential(*layers)

        # MLP after pooling
        layers = []
        for i in range(len(self.mlp2_dim) - 1):
            layers.append(nn.Linear(self.mlp2_dim[i], self.mlp2_dim[i + 1]))
            layers.append(nn.BatchNorm1d(self.mlp2_dim[i + 1]))
            if i < len(self.mlp2_dim) - 2:
                layers.append(nn.ReLU())
        self.mlp2 = nn.Sequential(*layers)

    def forward(self, input):
        """
        The input of the PSE is a tuple of tensors as yielded by the PixelSetData class:
          (Pixel-Set, Pixel-Mask) or ((Pixel-Set, Pixel-Mask), Extra-features)
        Pixel-Set : Batch_size x (Sequence length) x Channel x Number of pixels
        Pixel-Mask : Batch_size x (Sequence length) x Number of pixels
        Extra-features : Batch_size x (Sequence length) x Number of features
        If the input tensors have a temporal dimension, it will be combined with the batch dimension so that the
        complete sequences are processed at once. Then the temporal dimension is separated back to produce a tensor of
        shape Batch_size x Sequence length x Embedding dimension
        """
        a, b = input

        if len(a) == 2:
            out, mask = a
            extra = b
            if len(extra) == 2:
                extra, bm = extra
        else:
            out, mask = a, b

        if len(out.shape) == 4:
            # Combine batch and temporal dimensions in case of sequential input
            reshape_needed = True
            batch, temp = out.shape[:2]

            out = out.view(batch * temp, *out.shape[2:])
            mask = mask.view(batch * temp, -1)
            if self.with_extra:
                extra = extra.view(batch * temp, -1)
        else:
            reshape_needed = False

        out = self.mlp1(out)
        out = torch.cat(
            [pooling_methods[n](out, mask) for n in self.pooling.split("_")], dim=1
        )

        if self.with_extra:
            out = torch.cat([out, extra], dim=1)
        out = self.mlp2(out)

        if reshape_needed:
            out = out.view(batch, temp, -1)
        return out


class linlayer(nn.Module):
    def __init__(self, in_dim, out_dim):
        super(linlayer, self).__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim

        self.lin = nn.Linear(in_dim, out_dim)
        self.bn = nn.BatchNorm1d(out_dim)

    def forward(self, input):
        out = input.permute((0, 2, 1))  # to channel last
        out = self.lin(out)

        out = out.permute((0, 2, 1))  # to channel first
        out = self.bn(out)
        out = F.relu(out)

        return out


def masked_mean(x, mask):
    out = x.permute((1, 0, 2))
    out = out * mask
    out = out.sum(dim=-1) / mask.sum(dim=-1)
    out = out.permute((1, 0))
    return out


def masked_std(x, mask):
    m = masked_mean(x, mask)
    out = x.permute((2, 0, 1))
    out = out - m
    out = out.permute((2, 1, 0))

    out = out * mask
    d = mask.sum(dim=-1)
    d[d == 1] = 2

    out = (out**2).sum(dim=-1) / (d - 1)
    out = torch.sqrt(out + 10e-32)  # To ensure differentiability
    out = out.permute(1, 0)
    return out


def maximum(x, mask):
    return x.max(dim=-1)[0].squeeze()


def minimum(x, mask):
    return x.min(dim=-1)[0].squeeze()


pooling_methods = {
    "mean": masked_mean,
    "std": masked_std,
    "max": maximum,
    "min": minimum,
}

"""
Temporal Attention Encoder module
Credits:
The module is heavily inspired by the works of Vaswani et al. on self-attention and their pytorch implementation of
the Transformer served as code base for the present script.
paper: https://arxiv.org/abs/1706.03762
code: github.com/jadore801120/attention-is-all-you-need-pytorch
"""


class TemporalAttentionEncoder(nn.Module):
    def __init__(
        self,
        in_channels,
        n_head,
        d_k,
        d_model,
        n_neurons,
        dropout,
        T,
        len_max_seq,
        positions,
    ):
        """
        Sequence-to-embedding encoder.
        Args:
            in_channels (int): Number of channels of the input embeddings
            n_head (int): Number of attention heads
            d_k (int): Dimension of the key and query vectors
            n_neurons (list): Defines the dimensions of the successive feature spaces of the MLP that processes
                the concatenated outputs of the attention heads
            dropout (float): dropout
            T (int): Period to use for the positional encoding
            len_max_seq (int, optional): Maximum sequence length, used to pre-compute the positional encoding table
            positions (list, optional): List of temporal positions to use instead of position in the sequence
            d_model (int, optional): If specified, the input tensors will first processed by a fully connected layer
                to project them into a feature space of dimension d_model
        """

        super(TemporalAttentionEncoder, self).__init__()
        self.in_channels = in_channels
        self.positions = positions
        self.n_neurons = copy.deepcopy(n_neurons)

        self.name = "TAE_dk{}_{}Heads_{}_T{}_do{}".format(
            d_k, n_head, "|".join(list(map(str, self.n_neurons))), T, dropout
        )

        if positions is None:
            positions = len_max_seq + 1
        else:
            self.name += "_bespokePos"

        self.position_enc = nn.Embedding.from_pretrained(
            get_sinusoid_encoding_table(positions, self.in_channels, T=T), freeze=True
        )

        self.inlayernorm = nn.LayerNorm(self.in_channels)

        if d_model is not None:
            self.d_model = d_model
            self.inconv = nn.Sequential(
                nn.Conv1d(in_channels, d_model, 1), nn.LayerNorm([d_model, len_max_seq])
            )
            self.name += "_dmodel{}".format(d_model)
        else:
            self.d_model = in_channels
            self.inconv = None

        self.outlayernorm = nn.LayerNorm(self.d_model)

        self.attention_heads = MultiHeadAttention(
            n_head=n_head, d_k=d_k, d_in=self.d_model
        )

        assert self.n_neurons[0] == n_head * self.d_model
        assert self.n_neurons[-1] == self.d_model
        layers = []
        for i in range(len(self.n_neurons) - 1):
            layers.extend(
                [
                    nn.Linear(self.n_neurons[i], self.n_neurons[i + 1]),
                    nn.BatchNorm1d(self.n_neurons[i + 1]),
                    nn.ReLU(),
                ]
            )

        self.mlp = nn.Sequential(*layers)

        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        sz_b, seq_len, d = x.shape

        x = self.inlayernorm(x)

        if self.positions is None:
            src_pos = (
                torch.arange(1, seq_len + 1, dtype=torch.long)
                .expand(sz_b, seq_len)
                .to(x.device)
            )
        else:
            src_pos = (
                torch.arange(0, seq_len, dtype=torch.long)
                .expand(sz_b, seq_len)
                .to(x.device)
            )
        enc_output = x + self.position_enc(src_pos)

        if self.inconv is not None:
            enc_output = self.inconv(enc_output.permute(0, 2, 1)).permute(0, 2, 1)

        enc_output, attn = self.attention_heads(enc_output, enc_output, enc_output)

        enc_output = (
            enc_output.permute(1, 0, 2).contiguous().view(sz_b, -1)
        )  # Concatenate heads

        enc_output = self.outlayernorm(self.dropout(self.mlp(enc_output)))

        return enc_output


class MultiHeadAttention(nn.Module):
    """Multi-Head Attention module"""

    def __init__(self, n_head, d_k, d_in):
        super().__init__()
        self.n_head = n_head
        self.d_k = d_k
        self.d_in = d_in

        self.fc1_q = nn.Linear(d_in, n_head * d_k)
        nn.init.normal_(self.fc1_q.weight, mean=0, std=np.sqrt(2.0 / (d_k)))

        self.fc1_k = nn.Linear(d_in, n_head * d_k)
        nn.init.normal_(self.fc1_k.weight, mean=0, std=np.sqrt(2.0 / (d_k)))

        self.fc2 = nn.Sequential(
            nn.BatchNorm1d(n_head * d_k), nn.Linear(n_head * d_k, n_head * d_k)
        )

        self.attention = ScaledDotProductAttention(temperature=np.power(d_k, 0.5))

    def forward(self, q, k, v):
        d_k, d_in, n_head = self.d_k, self.d_in, self.n_head
        sz_b, seq_len, _ = q.size()

        q = self.fc1_q(q).view(sz_b, seq_len, n_head, d_k)
        q = q.mean(dim=1).squeeze()  # MEAN query
        q = self.fc2(q.view(sz_b, n_head * d_k)).view(sz_b, n_head, d_k)
        q = q.permute(1, 0, 2).contiguous().view(n_head * sz_b, d_k)

        k = self.fc1_k(k).view(sz_b, seq_len, n_head, d_k)
        k = k.permute(2, 0, 1, 3).contiguous().view(-1, seq_len, d_k)  # (n*b) x lk x dk

        v = v.repeat(n_head, 1, 1)  # (n*b) x lv x d_in

        output, attn = self.attention(q, k, v)

        output = output.view(n_head, sz_b, 1, d_in)
        output = output.squeeze(dim=2)

        return output, attn


class ScaledDotProductAttention(nn.Module):
    """Scaled Dot-Product Attention"""

    def __init__(self, temperature, attn_dropout=0.1):
        super().__init__()
        self.temperature = temperature
        self.dropout = nn.Dropout(attn_dropout)
        self.softmax = nn.Softmax(dim=2)

    def forward(self, q, k, v):
        attn = torch.matmul(q.unsqueeze(1), k.transpose(1, 2))
        attn = attn / self.temperature

        attn = self.softmax(attn)
        attn = self.dropout(attn)
        output = torch.matmul(attn, v)

        return output, attn


def get_sinusoid_encoding_table(positions, d_hid, T):
    """Sinusoid position encoding table
    positions: int or list of integer, if int range(positions)"""

    if isinstance(positions, int):
        positions = list(range(positions))

    def cal_angle(position, hid_idx):
        return position / np.power(T, 2 * (hid_idx // 2) / d_hid)

    def get_posi_angle_vec(position):
        return [cal_angle(position, hid_j) for hid_j in range(d_hid)]

    sinusoid_table = np.array([get_posi_angle_vec(pos_i) for pos_i in positions])

    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 2i
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  # dim 2i+1

    return torch.FloatTensor(sinusoid_table)


class FocalLoss(nn.Module):
    def __init__(self, gamma=0, alpha=None, size_average=True):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        if isinstance(alpha, (float, int)):
            self.alpha = torch.Tensor([alpha, 1 - alpha])
        if isinstance(alpha, list):
            self.alpha = torch.Tensor(alpha)
        self.size_average = size_average

    def forward(self, input, target):
        if input.dim() > 2:
            input = input.view(input.size(0), input.size(1), -1)  # N,C,H,W => N,C,H*W
            input = input.transpose(1, 2)  # N,C,H*W => N,H*W,C
            input = input.contiguous().view(-1, input.size(2))  # N,H*W,C => N*H*W,C
        target = target.view(-1, 1)
        logpt1 = F.log_softmax(input, dim=1)
        logpt2 = logpt1.gather(1, target.type(torch.int64))
        logpt3 = logpt2.view(-1)
        pt = Variable(logpt3.data.exp())

        if self.alpha is not None:
            if self.alpha.type() != input.data.type():
                self.alpha = self.alpha.type_as(input.data)
            at = self.alpha.gather(0, target.type(torch.int64).data.view(-1))
            logpt = logpt3 * Variable(at)

        loss = -1 * (1 - pt) ** self.gamma * logpt3
        if self.size_average:
            return loss.mean()
        else:
            return loss.sum()


class miou(Callback):
    def __init__(self, nclass):
        self.nclass = nclass

    def on_batch_end(self, last_output, last_target, **kwargs):
        self.y_pred = last_output
        self.y_true = last_target

    def on_epoch_end(self, last_metrics, **kwargs):
        y_pred = self.y_pred.argmax(dim=1)
        iou = 0
        n_classes = self.nclass
        n_observed = n_classes
        for i in range(n_classes):
            y_t = (self.y_true == i).int()
            y_p = (y_pred == i).int()

            inter = torch.sum(y_t * y_p)
            union = torch.sum((y_t + y_p > 0).int())

            if union == 0:
                n_observed -= 1
            else:
                iou += inter / union
        miou = iou / n_observed
        return add_metrics(last_metrics, miou.detach().cpu().numpy())


def mIou_new(y_true, y_pred, cls_list):
    """
    Mean Intersect over Union metric.
    Computes the one versus all IoU for each class and returns the average.
    Classes that do not appear in the provided set are not counted in the average.
    Args:
        y_true (1D-array): True labels
        y_pred (1D-array): Predicted labels
        n_classes (int): Total number of classes
    Returns:
        mean Iou (float)
    """
    iou = 0
    n_observed = len(cls_list)
    for i in cls_list:
        y_t = (np.array(y_true) == i).astype(int)
        y_p = (np.array(y_pred) == i).astype(int)

        inter = np.sum(y_t * y_p)
        union = np.sum((y_t + y_p > 0).astype(int))

        if union == 0:
            n_observed -= 1
        else:
            iou += inter / union
    return iou / n_observed


def weight_init(m):
    """
    Initializes a model's parameters.
    Credits to: https://gist.github.com/jeasinema
    Usage:
        model = Model()
        model.apply(weight_init)
    """
    if isinstance(m, nn.Conv1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.Conv3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose1d):
        init.normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose2d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.ConvTranspose3d):
        init.xavier_normal_(m.weight.data)
        if m.bias is not None:
            init.normal_(m.bias.data)
    elif isinstance(m, nn.BatchNorm1d):
        init.normal_(m.weight.data, mean=0, std=1)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm2d):
        init.normal_(m.weight.data, mean=0, std=1)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm3d):
        init.normal_(m.weight.data, mean=0, std=1)
        init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.Linear):
        init.xavier_normal_(m.weight.data)
        try:
            init.normal_(m.bias.data)
        except AttributeError:
            pass
    elif isinstance(m, nn.LSTM):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.LSTMCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRU):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)
    elif isinstance(m, nn.GRUCell):
        for param in m.parameters():
            if len(param.shape) >= 2:
                init.orthogonal_(param.data)
            else:
                init.normal_(param.data)


class PseTae(nn.Module):
    """
    Pixel-Set encoder + Temporal Attention Encoder sequence classifier
    """

    def __init__(self, input_dim, len_max_seq, positions, out_class, **kwargs):
        super(PseTae, self).__init__()

        mlp1 = kwargs.get("mlp1", [32, 64])
        mlp1inp = mlp1.copy()
        mlp1inp.insert(0, input_dim)

        pooling = kwargs.get("pooling", "mean_std")
        mlp2 = kwargs.get("mlp2", [128, 128])

        if pooling not in ["mean", "std", "max", "min"]:
            if mlp2[0] != mlp1[-1] * 2:
                raise Exception("MLP2 input should be double of MLP1 output")
        else:
            del mlp2[0]
            mlp2.insert(0, mlp1[1])

        n_head = kwargs.get("n_head", 4)
        d_k = kwargs.get("d_k", 32)

        mlp3 = []
        mlp3.extend([mlp2[1]] * 2)
        mlp3.insert(0, mlp2[1] * n_head)

        dropout = kwargs.get("dropout", 0.2)

        T = kwargs.get("T", 1000)

        mlp4 = kwargs.get("mlp4", [64, 32])
        mlp4inp = mlp4.copy()
        mlp4inp.insert(0, mlp3[-1])
        mlp4inp.append(out_class)

        d_model = int(mlp3[0] / n_head)

        self.spatial_encoder = PixelSetEncoder(
            input_dim,
            mlp1=mlp1inp,
            pooling=pooling,
            mlp2=mlp2,
            with_extra=False,
            extra_size=4,
        )
        self.temporal_encoder = TemporalAttentionEncoder(
            in_channels=mlp2[-1],
            n_head=n_head,
            d_k=d_k,
            d_model=d_model,
            n_neurons=mlp3,
            dropout=dropout,
            T=T,
            len_max_seq=len_max_seq,
            positions=positions,
        )
        self.decoder = get_decoder(mlp4inp)
        self.name = "_".join([self.spatial_encoder.name, self.temporal_encoder.name])

    def forward(self, input, target):
        """
        Args:
           input(tuple): (Pixel-Set, Pixel-Mask) or ((Pixel-Set, Pixel-Mask), Extra-features)
           Pixel-Set : Batch_size x Sequence length x Channel x Number of pixels
           Pixel-Mask : Batch_size x Sequence length x Number of pixels
           Extra-features : Batch_size x Sequence length x Number of features
        """
        out = self.spatial_encoder((input, target))
        out = self.temporal_encoder(out)
        out = self.decoder(out)
        return out

    def param_ratio(self):
        total = get_ntrainparams(self)
        s = get_ntrainparams(self.spatial_encoder)
        t = get_ntrainparams(self.temporal_encoder)
        c = get_ntrainparams(self.decoder)

        print("TOTAL TRAINABLE PARAMETERS : {}".format(total))
        print(
            "RATIOS: Spatial {:5.1f}% , Temporal {:5.1f}% , Classifier {:5.1f}%".format(
                s / total * 100, t / total * 100, c / total * 100
            )
        )


def get_ntrainparams(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def model_eval(data, model, class_dict, convertmap):
    validarr = torch.cat([i[0][None, :, :, :] for i, j in data.valid_ds], axis=0)
    batch_size = data.batch_size
    labsarr = torch.stack([j for i, j in data.valid_ds])
    final_img = torch.moveaxis(validarr, 3, 1)[:, :, :, :, None]
    img_arr = torch.reshape(
        final_img,
        (
            final_img.shape[0] * final_img.shape[1],
            final_img.shape[2],
            final_img.shape[3],
            1,
        ),
    ).to(model._device)
    final_labs = [validarr.shape[3] * [labsarr[l]] for l in range(validarr.shape[0])]
    final_labs = torch.stack(sum(final_labs, [])).to(model._device)
    divided = DataLoader(img_arr, batch_size=batch_size, pin_memory=False)
    prediction = []
    for i in progress_bar(divided):
        sim = torch.ones(i.shape[0], i.shape[1], 1).to(model._device)
        model.eval()
        with torch.no_grad():
            pred = model(i, sim)
        prediction.append(pred.argmax(dim=1))
    if convertmap:
        prediction = np.array(
            [convertmap.get(item, item) for item in torch.cat(prediction).cpu().numpy()]
        )
        final_labs = np.array(
            [convertmap.get(item, item) for item in final_labs.cpu().numpy()]
        )

    preds = np.array([class_dict.get(item, item) for item in prediction])
    trues = np.array([class_dict.get(item, item) for item in final_labs])
    mats = confusion_matrix_analysis(confusion_matrix(preds, trues), class_dict)
    miou = mIou_new(preds, trues, list(class_dict.values()))
    return mats, miou


def confusion_matrix_analysis(mat, cls_dict):
    """
    This method computes all the performance metrics from the confusion matrix.
    In addition to overall accuracy, the precision, recall, f-score and IoU for
    each class is computed.The class-wise metrics are averaged to provide overall
    indicators in two ways (MICRO and MACRO average).
    Args:
        mat (array): confusion matrix
    Returns:
        per_class (dict) : per class metrics
        overall (dict): overall metrics
    """
    TP = 0
    FP = 0
    FN = 0

    per_class = {}
    zero_divide = lambda n, d: 0 if n == 0 or d == 0 else n / d

    for j in range(mat.shape[0]):
        d = {}
        tp = np.sum(mat[j, j])
        fp = np.sum(mat[:, j]) - tp
        fn = np.sum(mat[j, :]) - tp

        d["IoU"] = zero_divide(tp, tp + fp + fn)
        d["Precision"] = zero_divide(tp, tp + fp)
        d["Recall"] = zero_divide(tp, tp + fn)
        d["F1-score"] = zero_divide(2 * tp, 2 * tp + fp + fn)

        per_class[str(list(cls_dict.values())[j])] = d

        TP += tp
        FP += fp
        FN += fn

    overall = {}
    overall["micro_IoU"] = TP / (TP + FP + FN)
    overall["micro_Precision"] = TP / (TP + FP)
    overall["micro_Recall"] = TP / (TP + FN)
    overall["micro_F1-score"] = 2 * TP / (2 * TP + FP + FN)

    macro = pd.DataFrame(per_class).transpose().mean()
    overall["MACRO_IoU"] = macro.loc["IoU"]
    overall["MACRO_Precision"] = macro.loc["Precision"]
    overall["MACRO_Recall"] = macro.loc["Recall"]
    overall["MACRO_F1-score"] = macro.loc["F1-score"]

    overall["Accuracy"] = np.sum(np.diag(mat)) / np.sum(mat)

    return per_class, overall



"""
MIT License

Copyright (c) 2020 Ziqiang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# based on https://github.com/ZQPei/deep_sort_pytorch
import os
import warnings
import traceback
from pathlib import Path

try:
    from numpy import isin, true_divide
    import torch
    import torch.nn as nn
    import torch.nn.functional as F
    from fastai.basic_train import LearnerCallback
    from fastai.basic_train import Learner
    from fastai.vision import imagenet_stats
    from fastai.vision.data import ImageDataBunch
    from fastai.vision.transform import ResizeMethod
    from fastai.metrics import accuracy
    from fastai.callback import Callback
    from fastai.torch_core import add_metrics
    from ._arcgis_model import _EmptyData

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False

REID_V1 = "reid_v1"
REID_V2 = "reid_v2"

DEFAULT_HEIGHT = 128
DEFAULT_WIDTH = 64
DEFAULT_CHANNELS = 3
DEFAULT_NUM_CLASSES = 4
DEFAULT_CLASS_MAPPING = {"obj_a": 1, "obj_b": 2, "obj_c": 3, "obj_d": 4}


class BasicBlock_v1(nn.Module):
    def __init__(self, c_in, c_out, is_downsample=False):
        super(BasicBlock_v1, self).__init__()

        self.is_downsample = is_downsample
        if is_downsample:
            self.conv1 = nn.Conv2d(c_in, c_out, 3, stride=2, padding=1, bias=False)
        else:
            self.conv1 = nn.Conv2d(c_in, c_out, 3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(c_out)
        self.relu = nn.ReLU(True)
        self.conv2 = nn.Conv2d(c_out, c_out, 3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(c_out)
        if is_downsample:
            self.downsample = nn.Sequential(
                nn.Conv2d(c_in, c_out, 1, stride=2, bias=False), nn.BatchNorm2d(c_out)
            )
        elif c_in != c_out:
            self.downsample = nn.Sequential(
                nn.Conv2d(c_in, c_out, 1, stride=1, bias=False), nn.BatchNorm2d(c_out)
            )
            self.is_downsample = True

    def forward(self, x):
        y = self.conv1(x)
        y = self.bn1(y)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.bn2(y)
        if self.is_downsample:
            x = self.downsample(x)
        return F.relu(x.add(y), True)


def make_layers_v1(c_in, c_out, repeat_times, is_downsample=False):
    blocks = []
    for i in range(repeat_times):
        if i == 0:
            blocks += [
                BasicBlock_v1(c_in, c_out, is_downsample=is_downsample),
            ]
        else:
            blocks += [
                BasicBlock_v1(c_out, c_out),
            ]
    return nn.Sequential(*blocks)


# TODO: Handle varied input sizes


class Net_v1(nn.Module):
    def __init__(self, num_classes=625, reid=False):
        super(Net_v1, self).__init__()

        self.img_shape = [3, 128, 64]
        self.conv = nn.Sequential(
            nn.Conv2d(3, 32, 3, stride=1, padding=1),
            nn.BatchNorm2d(32),
            nn.ELU(inplace=True),
            nn.Conv2d(32, 32, 3, stride=1, padding=1),
            nn.BatchNorm2d(32),
            nn.ELU(inplace=True),
            nn.MaxPool2d(3, 2, padding=1),
        )
        # 32 64 32
        self.layer1 = make_layers_v1(32, 32, 2, False)
        # 32 64 32
        self.layer2 = make_layers_v1(32, 64, 2, True)
        # 64 32 16
        self.layer3 = make_layers_v1(64, 128, 2, True)
        # 128 16 8
        self.dense = nn.Sequential(
            nn.Dropout(p=0.6),
            nn.Linear(128 * 16 * 8, 128),
            nn.BatchNorm1d(128),
            nn.ELU(inplace=True),
        )
        # 256 1 1
        self.reid = reid
        self.batch_norm = nn.BatchNorm1d(128)
        self.classifier = nn.Sequential(
            nn.Linear(128, num_classes),
        )

    def forward(self, x):
        x = self.conv(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = x.view(x.size(0), -1)
        if self.reid:
            x = self.dense[0](x)
            x = self.dense[1](x)
            x = x.div(x.norm(p=2, dim=1, keepdim=True))
            return x
        x = self.dense(x)
        # B x 128
        # classifier
        x = self.classifier(x)
        return x


class BasicBlock_v2(nn.Module):
    def __init__(self, c_in, c_out, is_downsample=False):
        super(BasicBlock_v2, self).__init__()

        self.is_downsample = is_downsample
        if is_downsample:
            self.conv1 = nn.Conv2d(c_in, c_out, 3, stride=2, padding=1, bias=False)
        else:
            self.conv1 = nn.Conv2d(c_in, c_out, 3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(c_out)
        self.relu = nn.ReLU(True)
        self.conv2 = nn.Conv2d(c_out, c_out, 3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(c_out)
        if is_downsample:
            self.downsample = nn.Sequential(
                nn.Conv2d(c_in, c_out, 1, stride=2, bias=False), nn.BatchNorm2d(c_out)
            )
        elif c_in != c_out:
            self.downsample = nn.Sequential(
                nn.Conv2d(c_in, c_out, 1, stride=1, bias=False), nn.BatchNorm2d(c_out)
            )
            self.is_downsample = True

    def forward(self, x):
        y = self.conv1(x)
        y = self.bn1(y)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.bn2(y)
        if self.is_downsample:
            x = self.downsample(x)
        return F.relu(x.add(y), True)


def make_layers_v2(c_in, c_out, repeat_times, is_downsample=False):
    blocks = []
    for i in range(repeat_times):
        if i == 0:
            blocks += [
                BasicBlock_v2(c_in, c_out, is_downsample=is_downsample),
            ]
        else:
            blocks += [
                BasicBlock_v2(c_out, c_out),
            ]
    return nn.Sequential(*blocks)


class Net_v2(nn.Module):
    def __init__(self, num_classes=184, reid=False):
        super(Net_v2, self).__init__()

        self.img_shape = [3, 128, 64]
        self.conv = nn.Sequential(
            nn.Conv2d(3, 64, 3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            # nn.Conv2d(32,32,3,stride=1,padding=1),
            # nn.BatchNorm2d(32),
            # nn.ReLU(inplace=True),
            nn.MaxPool2d(3, 2, padding=1),
        )
        # 32 64 32
        self.layer1 = make_layers_v2(64, 64, 2, False)
        # 32 64 32
        self.layer2 = make_layers_v2(64, 128, 2, True)
        # 64 32 16
        self.layer3 = make_layers_v2(128, 256, 2, True)
        # 128 16 8
        self.layer4 = make_layers_v2(256, 512, 2, True)
        # 256 8 4
        self.avgpool = nn.AvgPool2d((8, 4), 1)
        # 256 1 1
        self.reid = reid
        self.classifier = nn.Sequential(
            nn.Linear(512, 256),
            nn.BatchNorm1d(256),
            nn.ReLU(inplace=True),
            nn.Dropout(),
            nn.Linear(256, num_classes),
        )

    def forward(self, x):
        x = self.conv(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        # B x 128
        if self.reid:
            x = x.div(x.norm(p=2, dim=1, keepdim=True))
            return x
        # classifier
        x = self.classifier(x)
        return x


def build_opt_lr(trainable_params, **kwargs):
    optimizer = torch.optim.SGD(trainable_params, 0.1, momentum=0.9, weight_decay=5e-4)

    return optimizer


get_reid_loss = torch.nn.CrossEntropyLoss()


def _get_num_classes(data):
    if data is None or isinstance(data, _EmptyData):
        return DEFAULT_NUM_CLASSES
    return max(len(data.train_ds.y.classes), len(data.valid_ds.y.classes))


def get_default_backbone():
    return REID_V2


def get_default_imgsize():
    return (DEFAULT_HEIGHT, DEFAULT_WIDTH)


def get_model(num_classes, backbone, reid=False):
    if backbone is None:
        backbone = get_default_backbone()
    if backbone == REID_V1:
        return Net_v1(num_classes=num_classes, reid=reid)
    elif backbone == REID_V2:
        return Net_v2(num_classes, reid=reid)


def check_data_sanity(data, dataset_types=[]):
    try:
        if data is None:
            return False
        if isinstance(data, ImageDataBunch) and data._dataset_type in dataset_types:
            return True
        else:
            return False
    except Exception as e:
        return False


def _get_model_state(model_path, model=None):
    model_state = None
    if model_path is not None and os.path.isfile(model_path):
        if not torch.cuda.is_available():
            state = torch.load(model_path, map_location=lambda storage, loc: storage)
        else:
            device = torch.cuda.current_device()
            state = torch.load(
                model_path, map_location=lambda storage, loc: storage.cuda(device)
            )

            model_state = state
            if "model" in set(state.keys()):
                model_state = state["model"]
            elif isinstance(state, dict) and "net_dict" in state.keys():
                model_state = state["net_dict"]

    if model_state is None and model is not None and hasattr(model, "state_dict"):
        model_state = model.state_dict()

    return model_state


def load_for_prediction(model_path, num_classes, backbone, model):
    model_state = _get_model_state(model_path, model)
    model = get_model(num_classes, backbone, reid=True)
    if model_state is not None:
        model.load_state_dict(model_state)
    else:
        model = None

    return model


def _check_shape(data_in, img_shape):
    try:
        for i, data in enumerate(data_in):
            x = data[0][0]
            c_in = x.size()[0]
            h_in = x.size()[1]
            w_in = x.size()[2]

            if [c_in, h_in, w_in] != img_shape:
                return False
    except BaseException:
        return True
    return True


def _check_data_shape(data, img_shape):
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        if data is not None and isinstance(data, ImageDataBunch):
            is_valid = True
            if data.train_dl is not None:
                is_valid = is_valid and _check_shape(data.train_dl, img_shape)

            if data.valid_dl is not None:
                is_valid = is_valid and _check_shape(data.valid_dl, img_shape)

            if data.test_dl is not None:
                is_valid = is_valid and _check_shape(data.test_dl, img_shape)

            if not is_valid:
                raise Exception(
                    """\nInvalid input data shape.
                        DeepSort only supports input of the form
                        (channels=3, height=128, width=64)\n"""
                )


def get_fake_data():
    train_tfms = []
    val_tfms = []
    transforms = (train_tfms, val_tfms)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        kwargs_transforms = {}
        kwargs_transforms["size"] = get_default_imgsize()
        kwargs_transforms["resize_method"] = ResizeMethod.SQUISH
        data = ImageDataBunch.single_from_classes(
            Path(os.getcwd()),
            sorted(list(DEFAULT_CLASS_MAPPING.values())),
            ds_tfms=transforms,
        ).normalize(imagenet_stats)

    data.class_mapping = DEFAULT_CLASS_MAPPING
    data.classes = list(data.class_mapping.values())
    data._is_empty = True
    data._dataset_type = "Imagenet"
    data.resize_to = get_default_imgsize()

    return data


def reid_accuracy(input, targs):
    correct = input.max(dim=1)[1].eq(targs).sum().item()
    count = targs.size(0)

    return correct, count


class Accuracy(Callback):
    "Wrap a `func` in a callback for metrics computation."

    def __init__(self, func):
        # If func has a __name__ use this one else it should be a partial
        name = func.__name__ if hasattr(func, "__name__") else func.func.__name__
        self.func, self.name = func, name

    def on_epoch_begin(self, **kwargs):
        "Set the inner value to 0."
        self.correct, self.count = 0.0, 0

    def on_batch_end(self, last_output, last_target, **kwargs):
        "Update metric computation with `last_output` and `last_target`."
        if not isinstance(last_target, (tuple, list)):
            last_target = [last_target]
        correct, count = self.func(last_output, *last_target)
        self.correct += correct
        self.count += count

    def on_epoch_end(self, last_metrics, **kwargs):
        "Set the final result in `last_metrics`."
        return add_metrics(last_metrics, (1.0 * self.correct) / self.count)


def _get_metrics():
    return Accuracy(reid_accuracy)


def get_learner(data=None, num_classes=None, backbone=None, device=torch.device("cpu")):
    learn = None
    if num_classes is None:
        num_classes = _get_num_classes(data)

    model = get_model(num_classes, backbone)
    if model is None:
        return learn

    _check_data_shape(data, model.img_shape)

    model = model.to(device)
    metrics = _get_metrics()
    if data is not None:
        learn = Learner(
            data=data,
            model=model,
            loss_func=get_reid_loss,
            opt_func=build_opt_lr,
            metrics=metrics,
        )
    return learn



# Referred from https://github.com/tcapelle/timeseries_fastai

import torch
import torch.nn as nn
from .layers import *


def shortcut(c_in, c_out):
    return nn.Sequential(
        *[nn.Conv1d(c_in, c_out, kernel_size=1), nn.BatchNorm1d(c_out)]
    )


class _TSInceptionBlock(nn.Module):
    def __init__(self, input, nb_filters=32, ks=40, bottleneck=32):
        super().__init__()
        self.bottleneck = (
            nn.Conv1d(input, bottleneck, 1) if bottleneck and input > 1 else noop
        )

        kss = [ks // (2**i) for i in range(3)]
        kss = [ksi if ksi % 2 != 0 else ksi - 1 for ksi in kss]

        self.conv_layers = nn.ModuleList(
            [
                nn.Conv1d(
                    bottleneck if (bottleneck > 1 and input > 1) else input,
                    nb_filters,
                    kernel_size=ks,
                    padding=ks // 2,
                )
                for ks in kss
            ]
        )

        self.maxpool = nn.MaxPool1d(3, stride=1, padding=1)
        self.conv = nn.Conv1d(input, nb_filters, kernel_size=1)
        self.bn = nn.BatchNorm1d(nb_filters * 4)
        self.act = nn.ReLU()

    def forward(self, x):
        input_tensor = x
        x = self.bottleneck(input_tensor)
        layers_output = None
        for layer in self.conv_layers:
            output = layer(x)

            if layers_output is None:
                layers_output = output
            else:
                layers_output = torch.cat((layers_output, output), 1)

        mp = self.conv(self.maxpool(input_tensor))
        inc_out = torch.cat((layers_output, mp), 1)
        return self.act(self.bn(inc_out))


class _TSInceptionTime(nn.Module):
    def __init__(self, input, output, ks=40, depth=6, bottleneck=32, nb_filters=32):
        super().__init__()

        inception_layers = []
        residual_layers = []
        res = 0
        for d in range(depth):
            inception_layers.append(
                _TSInceptionBlock(
                    input if d == 0 else nb_filters * 4,
                    bottleneck=bottleneck if d > 0 else 0,
                    ks=ks,
                    nb_filters=nb_filters,
                )
            )
            if d % 3 == 2:
                if res == 0:
                    residual_layers.append(shortcut(input, nb_filters * 4))
                else:
                    residual_layers.append(shortcut(nb_filters * 4, nb_filters * 4))

                res = res + 1
            else:
                residual_layers.append(None)

        self.depth = depth
        self.inception_layers = nn.ModuleList(inception_layers)
        self.residual_layers = nn.ModuleList(residual_layers)
        self.act_fn = nn.ReLU()
        self.avg_pooling = nn.AdaptiveAvgPool1d(1)
        self.linear = nn.Linear(nb_filters * 4, output)

    def forward(self, x):
        res = x
        for d, l in enumerate(range(self.depth)):
            x = self.inception_layers[d](x)
            if d % 3 == 2:
                res = self.residual_layers[d](res)
                x += res
                res = x
                x = self.act_fn(x)

        x = self.avg_pooling(x).squeeze(-1)
        x = self.linear(x)
        return x



from typing import Union, Dict, Optional, ClassVar
from dataclasses import field, dataclass

from arcgis.realtime import Velocity
from arcgis.realtime.velocity.feeds._feed_template import _FeedTemplate
from arcgis.realtime.velocity.feeds.geometry import (
    _HasGeometry,
    SingleFieldGeometry,
    XYZGeometry,
)
from arcgis.realtime.velocity.feeds.run_interval import RunInterval
from arcgis.realtime.velocity.feeds.time import (
    _HasTime,
    TimeInstant,
    TimeInterval,
)
from arcgis.realtime.velocity.http_authentication_type import (
    NoAuth,
    BasicAuth,
    CertificateAuth,
)
from arcgis.realtime.velocity.input.format import (
    EsriJsonFormat,
    GeoJsonFormat,
    JsonFormat,
    DelimitedFormat,
    XMLFormat,
    _format_from_config,
)


@dataclass
class HttpPoller(_FeedTemplate, _HasTime, _HasGeometry):
    """
    Poll an HTTP endpoint for event data. This data class can be used to define the feed configuration and to create
    the feed.

    ===================      ====================================================================
    **Parameter**             **Description**
    -------------------      --------------------------------------------------------------------
    label                    String. Unique label for this feed instance.
    -------------------      --------------------------------------------------------------------
    description              String. Feed description.
    -------------------      --------------------------------------------------------------------
    url                      String. URL of the HTTP endpoint providing data.
    -------------------      --------------------------------------------------------------------
    http_http_method         String. HTTP method. Options: GET or POST.
    -------------------      --------------------------------------------------------------------
    http_auth_type           [:class:`~arcgis.realtime.velocity.NoAuth`, :class:`~arcgis.realtime.velocity.BasicAuth`, :class:`~arcgis.realtime.velocity.CertificateAuth`,OAuth]. An instance that contains the
                             authentication information for this feed instance.
    -------------------      --------------------------------------------------------------------
    url_params               Dict[str, str]. A dictionary of URL param/value pairs that contains
                             HTTP params used to access the HTTP resource.
    -------------------      --------------------------------------------------------------------
    http_headers             Dict[str, str]. A Name-Value dictionary that contains HTTP headers
                             for connecting to the HTTP resource.
    -------------------      --------------------------------------------------------------------
    enable_long_polling      bool. The default is: False.
    ===================      ====================================================================

    =====================     ====================================================================
    **Optional Argument**     **Description**
    =====================     ====================================================================
    data_format               [:class:`~arcgis.realtime.velocity.input.EsriJsonFormat`, :class:`~arcgis.realtime.velocity.input.GeoJsonFormat`, :class:`~arcgis.realtime.velocity.input.DelimitedFormat`, :class:`~arcgis.realtime.velocity.input.JsonFormat`, :class:`~arcgis.realtime.velocity.input.XMLFormat`].
                              An instance that contains the data format
                              configuration for this feed. Configure only allowed formats.
                              If this is not set right during initialization, a format will be
                              auto-detected and set from a sample of the incoming data. This sample
                              will be fetched from the configuration provided so far in the init.
    ---------------------     --------------------------------------------------------------------
    track_id_field            String. Name of the field from the incoming data that should be set as
                              track ID.
    ---------------------     --------------------------------------------------------------------
    geometry                  [:class:`~arcgis.realtime.velocity.feeds.XYZGeometry`, :class:`~arcgis.realtime.velocity.feeds.SingleFieldGeometry`]. An instance of geometry configuration
                              that will be used to create geometry objects from the incoming data.
    ---------------------     --------------------------------------------------------------------
    time                      [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`]. An instance of time configuration that
                              will be used to create time information from the incoming data.
    ---------------------     --------------------------------------------------------------------
    run_interval              :class:`~arcgis.realtime.velocity.feeds.RunInterval`. An instance of the scheduler configuration. The default is:
                              RunInterval(cron_expression="0 * * ? * * *", timezone="America/Los_Angeles")
    =====================     ====================================================================

    :return: A dataclass with Http poller feed configuration.

    .. code-block:: python

        # Usage Example

        from arcgis.realtime.velocity.feeds import HttpPoller
        from arcgis.realtime.velocity.http_authentication_type import (
            NoAuth,
            BasicAuth,
            CertificateAuth,
        )
        arcgis.realtime.velocity.input.format import DelimitedFormat
        from arcgis.realtime.velocity.feeds.geometry import XYZGeometry, SingleFieldGeometry
        from arcgis.realtime.velocity.feeds.time import TimeInterval, TimeInstant
        from arcgis.realtime.velocity.feeds.run_interval import RunInterval

        name = "http_poller_feed_name"
        description = "http_poller_description_feed"
        url = "http_poller_url"
        http_auth = NoAuth()
        # http_auth = BasicAuth(username="username", password="password")
        # http_auth = CertificateAuth(pfx_file_http_location="http_auth_link", password="password")

        http_headers = {"Content-Type": "application/json"}
        url_params = {"f": "json"}

        http_poller = HttpPoller(
            label=name,
            description=description,
            url=url,
            http_method="GET",
            http_auth_type=http_auth,
            url_params=url_params,
            http_headers=http_headers,
            enable_long_polling=False,
            data_format=None
        )

        # Set track id field
        http_poller.set_track_id("track_id")

        # Set time field
        time = TimeInstant(time_field="time_field")
        http_poller.set_time_config(time=time)

        # Set geometry field
        geometry = XYZGeometry(
            x_field="x",
            y_field="y",
            wkid=4326
        )
        http_poller.set_geometry_config(geometry=geometry)

        # Set recurrence
        http_poller.run_interval = RunInterval(
            cron_expression="0 * * ? * * *", timezone="America/Los_Angeles"
        )

        # use velocity object to get the FeedsManager instance
        feeds = velocity.feeds

        # use the FeedsManager object to create a feed from this feed configuration
        http_poller_feed = feeds.create(http_poller)
        http_poller_feed.start()
        feeds.items

    """

    # fields that the user sets during init
    # HTTP Poller specific properties
    url: str
    http_method: str
    http_auth_type: Union[NoAuth, BasicAuth, CertificateAuth]
    url_params: Dict[str, str] = field(default_factory=dict)
    http_headers: Dict[str, str] = field(default_factory=dict)
    enable_long_polling: bool = field(default=False)

    # user can define these properties even after initialization
    data_format: Optional[
        Union[
            EsriJsonFormat,
            GeoJsonFormat,
            JsonFormat,
            DelimitedFormat,
            XMLFormat,
        ]
    ] = None
    # FeedTemplate properties
    track_id_field: Optional[str] = None
    # HasGeometry properties
    geometry: Optional[Union[XYZGeometry, SingleFieldGeometry]] = None
    # HasTime properties
    time: Optional[Union[TimeInstant, TimeInterval]] = None
    # scheduler
    run_interval: RunInterval = field(
        default=RunInterval(
            cron_expression="0 * * ? * * *", timezone="America/Los_Angeles"
        )
    )

    # FeedTemplate properties
    _name: ClassVar[str] = "http-poller"

    def __post_init__(self):
        if Velocity is None:
            return
        self._util = Velocity._util

        # validation of fields
        if self._util.is_valid(self.label) == False:
            raise ValueError(
                "Label should only contain alpha numeric, _ and space only"
            )
        elif self.http_method not in ("POST", "GET"):
            raise ValueError("http_post str can either be 'POST' or 'GET'.")

        # generate dictionary of this feed object's properties that will be used to query test-connection and
        # sample-messages Rest endpoint
        feed_properties = self._generate_feed_properties()

        test_connection = self._util.test_connection(
            input_type="feed", payload=feed_properties
        )
        # Test connection to make sure feed can fetch schema
        if test_connection is True:
            # test connection succeeded. Now try getting sample messages
            sample_payload = {
                "properties": {
                    "maxSamplesToCollect": 5,
                    "timeoutInMillis": 5000,
                }
            }
            self._dict_deep_merge(sample_payload, feed_properties)
            # Sample messages to fetch schema/fields
            sample_messages = self._util.sample_messages(
                input_type="feed", payload=sample_payload
            )

            if sample_messages["featureSchema"] is not None:
                # sample messages succeeded. Get Feature Schema from it
                self._set_fields(sample_messages["featureSchema"])

            # if Format was not specified by user, use the auto-detected format from the sample messages response as this feed object's format.
            if self.data_format is None:
                self.data_format = _format_from_config(sample_messages)

            print(
                "Feature Schema retrieved from the Feed:",
                sample_messages["featureSchema"],
            )

            # initiate actions for each of the following properties if it was set at init
            if self.track_id_field is not None:
                self.set_track_id(self.track_id_field)
            if self.geometry is not None:
                self.set_geometry_config(self.geometry)
            if self.time is not None:
                self.set_time_config(self.time)

    def _build(self) -> dict:
        feed_configuration = {
            "id": "",
            "label": self.label,
            "description": self.description,
            "feed": {**self._generate_schema_transformation()},
            **self.run_interval._build(),
            "properties": {"executable": True},
        }

        feed_properties = self._generate_feed_properties()
        self._dict_deep_merge(feed_configuration["feed"], feed_properties)
        print(feed_configuration)
        return feed_configuration

    def _generate_feed_properties(self) -> dict:
        # http headers
        if bool(self.http_headers):
            http_headers_properties = {f"{self._name}.headers": self.http_headers}
        else:
            http_headers_properties = {}
        # url params
        if bool(self.url_params):
            url_params_properties = {f"{self._name}.urlParameters": self.url_params}
        else:
            url_params_properties = {}

        # http authentication type
        auth_properties = self.http_auth_type._build(self._name)

        feed_properties = {
            "name": self._name,
            "properties": {
                f"{self._name}.url": self.url,
                f"{self._name}.httpMethod": self.http_method,
                f"{self._name}.isLongPolling": self.enable_long_polling,
                **auth_properties,
                **http_headers_properties,
                **url_params_properties,
            },
        }

        if self.data_format is not None:
            format_dict = self.data_format._build()
            self._dict_deep_merge(feed_properties, format_dict)

        return feed_properties



import os
import sys
import json
import arcgis
from arcgis.learn import ImageCaptioner

import numpy as np
from .util import normalize_batch

try:
    from fastai.vision import *
    import torch

    HAS_PYTORCH_FA = True

except Exception as e:
    HAS_PYTORCH_FA = False

imagenet_stats = ([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

imagenet_mean = 255 * np.array(imagenet_stats[0], dtype=np.float32)
imagenet_std = 255 * np.array(imagenet_stats[1], dtype=np.float32)


def norm(x, mean=imagenet_mean, std=imagenet_std):
    return (x - mean) / std


def denorm(x, mean=imagenet_mean, std=imagenet_std):
    return x * std + mean


class ChildObjectDetector:
    def initialize(self, model, model_as_file):
        if not HAS_PYTORCH_FA:
            raise Exception(
                'PyTorch (version 1.1.0 or above) and fast.ai (version 1.0.54 or above) libraries are not installed. Install PyTorch using "conda install pytorch=1.1.0 fastai=1.0.54".'
            )

        if model_as_file:
            with open(model, "r") as f:
                self.emd = json.load(f)
        else:
            self.emd = json.loads(model)
        import arcpy

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device("cuda")
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device("cpu")
            arcgis.env._processorType = "CPU"

        # Using arcgis.learn FeatureClassifer from_model function.
        self.cf = ImageCaptioner.from_model(emd_path=model)
        self._learnmodel = self.cf
        self.model = self.cf.learn.model.to(self.device)
        self.model.eval()

    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    "name": "beam_width",
                    "dataType": "numeric",
                    "value": 5,
                    "required": False,
                    "displayName": "The number of captions to consider.",
                    "description": "The number of captions to consider.",
                },
                {
                    "name": "max_length",
                    "dataType": "numeric",
                    "value": 20,
                    "required": False,
                    "displayName": "Maximum length of the caption.",
                    "description": "Maximum length of the caption.",
                },
            ]
        )
        return required_parameters

    def getConfiguration(self, **scalars):
        if "BatchSize" not in self.emd and "batch_size" not in scalars:
            self.batch_size = 1
        elif "BatchSize" not in self.emd and "batch_size" in scalars:
            self.batch_size = int(scalars["batch_size"])
        else:
            self.batch_size = int(self.emd["BatchSize"])

        self.beam_width = int(scalars.get("beam_width", 5))

        self.max_length = int(scalars.get("max_length", 20))

        return {
            # CropSizeFixed is a boolean value parameter (1 or 0) in the emd file, representing whether the size of
            # tile cropped around the feature is fixed or not.
            # 1 -- fixed tile size, crop fixed size tiles centered on the feature. The tile can be bigger or smaller
            # than the feature;
            # 0 -- Variable tile size, crop out the feature using the smallest fitting rectangle. This results in tiles
            # of varying size, both in x and y. the ImageWidth and ImageHeight in the emd file are still passed and used
            # as a maximum size. If the feature is bigger than the defined ImageWidth/ImageHeight, the tiles are cropped
            # the same way as in the fixed tile size option using the maximum size.
            "CropSizeFixed": int(self.emd["CropSizeFixed"]),
            # BlackenAroundFeature is a boolean value paramater (1 or 0) in the emd file, representing whether blacken
            # the pixels outside the feature in each image tile.
            # 1 -- Blacken
            # 0 -- Not blacken
            "BlackenAroundFeature": int(self.emd["BlackenAroundFeature"]),
            "extractBands": tuple(self.emd["ExtractBands"]),
            "tx": self.emd["ImageWidth"],
            "ty": self.emd["ImageHeight"],
            "batch_size": self.batch_size,
        }

    def vectorize(self, **pixelBlocks):
        # Get pixel blocks - tuple of 3-d rasters: ([bands,height,width],[bands,height.width],...)
        # Convert tuple to 4-d numpy array
        batch_images = np.asarray(pixelBlocks["rasters_pixels"])

        # Get the shape of the 4-d numpy array
        batch, bands, height, width = batch_images.shape

        rings = []
        labels, confidences = [], []

        # Normalize Image
        if "NormalizationStats" in self.emd:
            batch_images = normalize_batch(batch_images, self.emd)
        else:
            # Transpose the image dimensions to [batch, height, width, bands],
            # normalize and transpose back to [batch, bands, height, width]
            batch_images = norm(batch_images.transpose(0, 2, 3, 1)).transpose(
                0, 3, 1, 2
            )

        # Convert to torch tensor, set device and convert to float
        batch_images = torch.tensor(batch_images).to(self.device).float()

        _, labels, _ = self.cf.learn.model.sample(
            batch_images.to(self.device), self.beam_width, self.max_length
        )

        # Appending this ring for all the features in the batch
        rings = [
            [[0, 0], [width - 1, 0], [width - 1, height - 1], [0, height - 1]]
            for i in range(batch)
        ]

        return rings, labels



# Code from https://github.com/fastai/course-v3/blob/master/nbs/dl2/cyclegan_ws.ipynb, https://github.com/eriklindernoren/PyTorch-GAN & https://github.com/haofengac/MonoDepth-FPN-PyTorch

import torch
import random
import numpy as np
from fastai.vision import (
    nn,
    Callable,
    List,
    LearnerCallback,
    optim,
    ifnone,
    F,
    flatten_model,
    requires_grad,
    SmoothenValue,
    add_metrics,
)
from .._utils.cyclegan import (
    calculate_activation_statistics,
    calculate_frechet_distance,
)
from fastprogress.fastprogress import progress_bar
from .._utils.superres import psnr, ssim
from skimage.feature import match_template

import torch.nn as nn
import torch.nn.functional as F
import torch


def weights_init_normal(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        torch.nn.init.normal_(m.weight.data, 0.0, 0.02)
    elif classname.find("BatchNorm2d") != -1:
        torch.nn.init.normal_(m.weight.data, 1.0, 0.02)
        torch.nn.init.constant_(m.bias.data, 0.0)


##############################
#          Hybrid W-NET
##############################


class UNetDown(nn.Module):
    def __init__(self, in_size, out_size, normalize=True, dropout=0.0):
        super(UNetDown, self).__init__()
        layers = [nn.Conv2d(in_size, out_size, 4, 2, 1, bias=False)]
        if normalize:
            layers.append(nn.InstanceNorm2d(out_size))
        layers.append(nn.LeakyReLU(0.2))
        if dropout:
            layers.append(nn.Dropout(dropout))
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class UNetUp(nn.Module):
    def __init__(self, in_size, out_size, dropout=0.0, kernel=4, stride=2, padding=1):
        super(UNetUp, self).__init__()
        layers = [
            nn.ConvTranspose2d(in_size, out_size, kernel, stride, padding, bias=False),
            nn.InstanceNorm2d(out_size),
            nn.ReLU(inplace=True),
        ]
        if dropout:
            layers.append(nn.Dropout(dropout))

        self.model = nn.Sequential(*layers)

    def forward(self, x, *skip_input):
        x = self.model(x)
        if skip_input:
            x = torch.cat((x, skip_input[0], skip_input[1]), 1)
            return x
        else:
            return x


class GeneratorUNet(nn.Module):
    def __init__(self, in_channels=3, out_channels=3):
        super(GeneratorUNet, self).__init__()

        self.enc_down1 = UNetDown(in_channels, 64, normalize=False)
        self.enc_down2 = UNetDown(64, 128)
        self.enc_down3 = UNetDown(128, 256)
        self.enc_down4 = UNetDown(256, 512, dropout=0.5)
        self.enc_down5 = UNetDown(512, 512, dropout=0.5)
        self.enc_down6 = UNetDown(512, 512, dropout=0.5)
        self.enc_down7 = UNetDown(512, 512, dropout=0.5)
        self.enc_down8 = UNetDown(512, 512, normalize=False, dropout=0.5)

        self.up1 = UNetUp(1024, 1024, dropout=0.5, kernel=1, stride=1, padding=0)
        self.up2 = UNetUp(1024, 512, dropout=0.5)
        self.up3 = UNetUp(1536, 512, dropout=0.5)
        self.up4 = UNetUp(1536, 512, dropout=0.5)
        self.up5 = UNetUp(1536, 512, dropout=0.5)
        self.up6 = UNetUp(1536, 256)
        self.up7 = UNetUp(768, 128)
        self.up8 = UNetUp(384, 64)

        self.final = nn.Sequential(
            nn.Upsample(scale_factor=2),
            nn.ZeroPad2d((1, 0, 1, 0)),
            nn.Conv2d(192, out_channels, 4, padding=1),
            nn.Tanh(),
        )

    def forward(self, x1, x2):
        # U-Net generator with skip connections from encoder to decoder
        d1 = self.enc_down1(x1)
        d2 = self.enc_down2(d1)
        d3 = self.enc_down3(d2)
        d4 = self.enc_down4(d3)
        d5 = self.enc_down5(d4)
        d6 = self.enc_down6(d5)
        d7 = self.enc_down7(d6)
        d8 = self.enc_down8(d7)

        e1 = self.enc_down1(x2)
        e2 = self.enc_down2(e1)
        e3 = self.enc_down3(e2)
        e4 = self.enc_down4(e3)
        e5 = self.enc_down5(e4)
        e6 = self.enc_down6(e5)
        e7 = self.enc_down7(e6)
        e8 = self.enc_down8(e7)

        conc = torch.cat((d8, e8), 1)

        u1 = self.up1(conc)
        u2 = self.up2(u1, d7, e7)
        u3 = self.up3(u2, d6, e6)
        u4 = self.up4(u3, d5, e5)
        u5 = self.up5(u4, d4, e4)
        u6 = self.up6(u5, d3, e3)
        u7 = self.up7(u6, d2, e2)
        u8 = self.up8(u7, d1, e1)

        return self.final(u8)


##############################
#        Discriminator
##############################


class discriminator_block(nn.Module):
    def __init__(self, in_filters, out_filters, normalization=True):
        super(discriminator_block, self).__init__()
        """Returns downsampling layers of each discriminator block"""
        layers = [nn.Conv2d(in_filters, out_filters, 4, stride=2, padding=1)]
        if normalization:
            layers.append(nn.BatchNorm2d(out_filters))
        layers.append(nn.LeakyReLU(0.2, inplace=True))
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class Discriminator(nn.Module):
    def __init__(self, in_channels=3):
        super(Discriminator, self).__init__()

        self.disc_blk_1 = discriminator_block(in_channels * 2, 64, normalization=False)
        self.disc_blk_2 = discriminator_block(64, 128)
        self.disc_blk_3 = discriminator_block(128, 256)
        self.disc_blk_4 = discriminator_block(256, 512)
        self.final = nn.Sequential(
            nn.ZeroPad2d((1, 0, 1, 0)),
            nn.Conv2d(512, 1, 4, padding=1, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, img_A, img_B):
        # Concatenate image and condition image by channels to produce input
        img_input = torch.cat((img_A, img_B), 1)
        disc_1 = self.disc_blk_1(img_input)
        disc_2 = self.disc_blk_2(disc_1)
        disc_3 = self.disc_blk_3(disc_2)
        disc_4 = self.disc_blk_4(disc_3)

        return self.final(disc_4)


class WNet_cGAN(nn.Module):
    def __init__(
        self,
        ch_in: int,
        ch_out: int,
        n_features: int = 64,
        disc_layers: int = 3,
        gen_blocks: int = 8,
        lsgan: bool = False,
        drop: float = 0.0,
        norm_layer: nn.Module = None,
    ):
        super().__init__()

        self.D = Discriminator(ch_in)
        self.G = GeneratorUNet(ch_in, ch_out)
        self.arcgis_results = False

    def forward(self, real_A, real_B, real_C):
        if real_A.shape[0] == 1 or real_B.shape[0] == 1 or real_C.shape[0] == 1:
            real_A, real_B, real_C = (
                torch.tensor([real_A.cpu().numpy()[0, :, :, :]] * 2).cuda(),
                torch.tensor([real_B.cpu().numpy()[0, :, :, :]] * 2).cuda(),
                torch.tensor([real_C.cpu().numpy()[0, :, :, :]] * 2).cuda(),
            )
        fake_C = self.G(real_A, real_B)
        if real_A.shape[0] == 1 or real_B.shape[0] == 1:
            fake_C = fake_C[None, 0, :, :, :]
        if self.training:
            self.arcgis_results = False
        if self.arcgis_results:
            return torch.cat([fake_C[:, None], fake_C[:, None]], 1)
        return [fake_C]


class AdaptiveLoss(nn.Module):
    def __init__(self, crit):
        super().__init__()
        self.crit = crit

    def forward(self, output, target: bool, **kwargs):
        targ = (
            output.new_ones(*output.size())
            if target
            else output.new_zeros(*output.size())
        )
        return self.crit(output, targ, **kwargs)


class Adaptivel2Loss(nn.Module):
    def __init__(self, crit1):
        super().__init__()
        self.crit1 = crit1

    def forward(self, output, target: bool, **kwargs):
        targ = (
            output.new_ones(*output.size())
            if target
            else output.new_zeros(*output.size())
        )
        return self.crit1(output, targ, **kwargs)


class RMSE_log(nn.Module):
    def __init__(self):
        super(RMSE_log, self).__init__()

    def forward(self, fake, real):
        eps = 1e-7
        real = F.relu(real)
        fake = F.relu(fake)
        loss = torch.sqrt(
            torch.mean(torch.abs(torch.log(real + eps) - torch.log(fake + eps)) ** 2)
        )
        return loss


class NormalLoss(nn.Module):
    def __init__(self):
        super(NormalLoss, self).__init__()

    def imgrad(self, img):
        img = torch.mean(img, 1, True)
        fx = np.array([[1, 0, -1], [2, 0, -2], [1, 0, -1]])
        conv1 = nn.Conv2d(1, 1, kernel_size=3, stride=1, padding=1, bias=False)
        weight = torch.from_numpy(fx).float().unsqueeze(0).unsqueeze(0)
        if img.is_cuda:
            weight = weight.cuda()
        conv1.weight = nn.Parameter(weight)
        grad_x = conv1(img)

        fy = np.array([[1, 2, 1], [0, 0, 0], [-1, -2, -1]])
        conv2 = nn.Conv2d(1, 1, kernel_size=3, stride=1, padding=1, bias=False)
        weight = torch.from_numpy(fy).float().unsqueeze(0).unsqueeze(0)
        if img.is_cuda:
            weight = weight.cuda()
        conv2.weight = nn.Parameter(weight)
        grad_y = conv2(img)
        return grad_y, grad_x

    def imgrad_yx(self, img):
        N, C, _, _ = img.size()
        grad_y, grad_x = self.imgrad(img)
        return torch.cat((grad_y.view(N, C, -1), grad_x.view(N, C, -1)), dim=1)

    def forward(self, grad_fake, grad_real):
        c, d = (
            self.imgrad_yx((grad_fake[:, 0, :, :])[:, None, :, :]),
            self.imgrad_yx((grad_real[:, 0, :, :])[:, None, :, :]),
        )

        e = (c[:, :, None, :] @ d[:, :, :, None]).squeeze(-1).squeeze(-1)

        fake_norm = torch.sqrt(torch.sum(c**2, dim=-1))
        real_norm = torch.sqrt(torch.sum(d**2, dim=-1))

        return 1 - torch.mean(e / (fake_norm * real_norm))


class get_activation_from_layers:
    def __init__(self, name):
        self.name = name

    def get_activation(self):
        activation = {}

        def hook(model, input, output):
            activation[self.name] = output.detach()

        return hook, activation


class WNetcGANLoss(nn.Module):
    def __init__(
        self,
        cgan: nn.Module,
        lambda_A: float = 100.0,
        lambda_B: float = 10,
        lambda_idt: float = 0.5,
        gamma_g: float = 10.0,
        lsgan: bool = False,
    ):
        super().__init__()
        self.cgan, self.l_A, self.l_B, self.l_idt, self.gam_g = (
            cgan,
            lambda_A,
            lambda_B,
            lambda_idt,
            gamma_g,
        )
        self.crit = AdaptiveLoss(
            F.mse_loss if lsgan else F.binary_cross_entropy_with_logits
        )
        self.crit1 = Adaptivel2Loss(F.mse_loss)
        self.norm_loss = NormalLoss()
        self.rmse_log = RMSE_log()

    def set_input(self, input):
        self.real_A, self.real_B, self.real_C = input
        if (
            self.real_A.shape[0] == 1
            or self.real_B.shape[0] == 1
            or self.real_C.shape[0] == 1
        ):
            self.real_A, self.real_B, self.real_C = (
                torch.tensor([self.real_A.cpu().numpy()[0, :, :, :]] * 2).cuda(),
                torch.tensor([self.real_B.cpu().numpy()[0, :, :, :]] * 2).cuda(),
                torch.tensor([self.real_C.cpu().numpy()[0, :, :, :]] * 2).cuda(),
            )

    def forward(self, output, target):
        # fake_C = (output)[:,0,:,:,:]
        fake_C = (output)[0]
        self.gen_loss = self.crit(self.cgan.D(self.real_A, fake_C), True)
        self.l1_loss = F.l1_loss(self.real_C, fake_C)
        self.rmse_log_loss = self.rmse_log(self.real_C, fake_C)
        self.norm_vec_loss = self.norm_loss(fake_C, self.real_C)
        return (
            self.gen_loss
            + (self.l_A * self.l1_loss)
            + (self.l_B * self.norm_vec_loss)
            + self.rmse_log_loss
        )


class WNetcGANTrainer(LearnerCallback):
    _order = -20  # Need to run before the Recorder

    def _set_trainable(self, D=False):
        gen = not D
        # requires_grad(self.learn.model.G, gen)
        # requires_grad(self.learn.model.D, D)
        if not gen:
            self.opt_D.lr, self.opt_D.mom = self.learn.opt.lr, self.learn.opt.mom
            self.opt_D.wd, self.opt_D.beta = self.learn.opt.wd, self.learn.opt.beta

    def on_train_begin(self, **kwargs):
        self.G = self.learn.model.G
        self.D = self.learn.model.D
        self.crit = self.learn.loss_func.crit
        self.crit1 = self.learn.loss_func.crit1

        if not getattr(self, "opt_G", None):
            self.opt_G = self.learn.opt.new([nn.Sequential(*flatten_model(self.G))])
        else:
            self.opt_G.lr, self.opt_G.wd = self.opt.lr, self.opt.wd
            self.opt_G.mom, self.opt_G.beta = self.opt.mom, self.opt.beta

        if not getattr(self, "opt_D", None):
            self.opt_D = self.learn.opt.new([nn.Sequential(*flatten_model(self.D))])

        self.learn.opt.opt = self.opt_G.opt
        self._set_trainable()
        self.gen_smter, self.l1_smter, self.norm_vec_smter = (
            SmoothenValue(0.98),
            SmoothenValue(0.98),
            SmoothenValue(0.98),
        )
        self.d_smter = SmoothenValue(0.98)
        self.recorder.add_metric_names(
            ["gen_loss", "l1_loss", "norm_vec_loss", "D_loss"]
        )

    def on_batch_begin(self, last_input, **kwargs):
        self.learn.loss_func.set_input(last_input)

    def on_backward_begin(self, **kwargs):
        self.l1_smter.add_value(self.loss_func.l1_loss.detach().cpu())
        self.gen_smter.add_value(self.loss_func.gen_loss.detach().cpu())
        self.norm_vec_smter.add_value(self.loss_func.norm_vec_loss)

    def on_batch_end(self, last_input, last_output, **kwargs):
        self.G.zero_grad()
        fake_C = last_output[0].detach()
        real_A, real_B, real_C = last_input
        if real_A.shape[0] == 1 or real_B.shape[0] == 1 or real_C.shape[0] == 1:
            real_A, real_B, real_C = (
                torch.tensor([real_A.cpu().numpy()[0, :, :, :]] * 2).cuda(),
                torch.tensor([real_B.cpu().numpy()[0, :, :, :]] * 2).cuda(),
                torch.tensor([real_C.cpu().numpy()[0, :, :, :]] * 2).cuda(),
            )
        self._set_trainable(D=True)

        self.D.zero_grad()

        if random.choice([0, 1]) < 0.5:
            loss_D = self.crit(self.D(real_A, real_C), True)
        else:
            loss_D = self.crit(self.D(real_A, fake_C), False)

        self.d_smter.add_value(loss_D.detach().cpu())
        if self.learn.model.training == True:
            loss_D.backward()

        self.opt_D.step()

        self._set_trainable()

    def on_epoch_end(self, last_metrics, **kwargs):
        return add_metrics(
            last_metrics,
            [
                s.smooth
                for s in [
                    self.gen_smter,
                    self.l1_smter,
                    self.norm_vec_smter,
                    self.d_smter,
                ]
            ],
        )


def compute_metrics(model, dl, show_progress):
    avg_psnr = 0
    avg_ssim = 0
    avg_ncc = 0
    model.learn.model.eval()
    with torch.no_grad():
        for input, target in progress_bar(dl, display=False):
            if (
                input[0].shape[0] != 1
                or input[1].shape[0] != 1
                or input[2].shape[0] != 1
            ):
                prediction = model.learn.model(input[0], input[1], input[2])
                if isinstance(prediction, list):
                    prediction = prediction[0]
                else:
                    prediction = prediction[:, 0, :, :, :]
                avg_ncc += match_template(
                    prediction.cpu().numpy()[:, :, 0], input[2].cpu().numpy()[:, :, 0]
                )
                avg_psnr += psnr(prediction, input[2])
                avg_ssim += ssim(prediction, input[2])
    return avg_psnr / len(dl), avg_ssim.item() / len(dl), float(avg_ncc / len(dl))
